import logging
from dataclasses import dataclass

from nlu import Licenses
from nlu.pipe.component_resolution import resolve_feature
from nlu.pipe.nlu_component import NluComponent
from nlu.pipe.pipeline import NLUPipeline
from nlu.pipe.utils.component_utils import ComponentUtils
from nlu.pipe.utils.pipe_utils import PipeUtils
from nlu.pipe.utils.resolution.storage_ref_utils import StorageRefUtils
from nlu.universe.component_universes import jsl_id_to_empty_component
from nlu.universe.feature_node_ids import NLP_NODE_IDS, NLP_HC_NODE_IDS
from nlu.universe.feature_universes import NLP_FEATURES
from nlu.universe.logic_universes import AnnoTypes

logger = logging.getLogger('nlu')


@dataclass
class StorageRefConversionResolutionData:
    """Hold information that can be used to resolve to a NLU component_to_resolve, which satisfies the storage ref demands."""
    storage_ref: str  # storage ref a resolver component_to_resolve should have
    component_candidate: NluComponent  # from which NLU component_to_resolve should the converter feed
    type: str  # what kind of conversion, either word2chunk or word2sentence


class PipelineCompleter:
    '''
        Pass a list of NLU components to the pipeline (or a NLU pipeline)
        For every component_to_resolve, it checks if all requirements are met.
        It checks and fixes the following issues  for a list of components:
        1. Missing Features / component_to_resolve requirements
        2. Bad order of components (which will cause missing features exception)
        3. Check Feature names in the output
        4. Check weather pipeline needs to be fitted
    '''

    @staticmethod
    def check_if_storage_ref_is_satisfied_or_get_conversion_candidate(component_to_check: NluComponent,
                                                                      pipe: NLUPipeline,
                                                                      storage_ref_to_find: str):
        """Check if any other component_to_resolve in the pipeline has same storage ref as the input component_to_resolve.
        Returns 1. If there is a candidate, but it has different level, it will be returned as candidate
        If first condition is not satisfied, consults the namespace.storage_ref_2_nlp_ref
        """
        # If there is just 1 component_to_resolve, there is nothing to check
        if len(pipe.components) == 1:
            return False, None
        conversion_candidate = None
        conversion_type = "no_conversion"
        logger.info(f'checking for storage={storage_ref_to_find} is available in component_list..')
        for c in pipe.components:
            if component_to_check.name != c.name:
                if StorageRefUtils.has_storage_ref(c):
                    if StorageRefUtils.extract_storage_ref(c) == storage_ref_to_find:
                        # Both components have Different Names AND their Storage Ref Matches up AND they both take in tokens -> Match
                        if NLP_FEATURES.TOKEN in component_to_check.in_types and c.type == AnnoTypes.TOKEN_EMBEDDING:
                            logger.info(f'Word Embedding Match found = {c.name}')
                            return False, None

                        # Since document and be substituted for sentence
                        # and vice versa if either of them matches up we have a match
                        if NLP_FEATURES.SENTENCE_EMBEDDINGS in component_to_check.in_types and \
                                c.type == AnnoTypes.DOCUMENT_EMBEDDING:
                            logger.info(f'Sentence Embedding Match found = {c.name}')
                            return False, None

                        # component_to_check requires Sentence_embedding
                        # but the Matching Storage_ref component_to_resolve takes in Token
                        #   -> Convert the Output of the Match to SentenceLevel
                        #   and feed the component_to_check to the new component_to_resolve
                        if NLP_FEATURES.SENTENCE_EMBEDDINGS in component_to_check.in_types \
                                and c.type == AnnoTypes.TOKEN_EMBEDDING:
                            logger.info(f'Sentence Embedding Conversion Candidate found={c.name}')
                            conversion_type = 'word2sentence'
                            conversion_candidate = c

                        # analogous case as above for chunk
                        if NLP_FEATURES.CHUNK_EMBEDDINGS in component_to_check.in_types and c.type == AnnoTypes.TOKEN_EMBEDDING:
                            logger.info(f'Sentence Embedding Conversion Candidate found={c.name}')
                            conversion_type = 'word2chunk'
                            conversion_candidate = c

        logger.info(f'No matching storage ref found')
        return True, StorageRefConversionResolutionData(storage_ref_to_find, conversion_candidate, conversion_type)

    @staticmethod
    def extract_required_features_refless_from_pipe(pipe: NLUPipeline):
        """Extract provided features from component_list, which have no storage ref"""
        provided_features_no_ref = []
        for c in pipe.components:
            if c.loaded_from_pretrained_pipe:
                continue
            for feat in c.in_types:
                if 'embed' not in feat: provided_features_no_ref.append(feat)
        return ComponentUtils.clean_irrelevant_features(provided_features_no_ref)

    @staticmethod
    def extract_provided_features_refless_from_pipe(pipe: NLUPipeline):
        """Extract provided features from component_list, which have no storage ref"""
        provided_features_no_ref = []
        for c in pipe.components:
            for feat in c.out_types:
                if 'embed' not in feat: provided_features_no_ref.append(feat)
        return ComponentUtils.clean_irrelevant_features(provided_features_no_ref)

    @staticmethod
    def extract_provided_features_ref_from_pipe(pipe: NLUPipeline):
        """Extract provided features from component_list, which have  storage ref.
        """
        provided_features_ref = []
        for c in pipe.components:
            for feat in c.out_types:
                if 'embed' in feat:
                    if '@' not in feat:
                        provided_features_ref.append(feat + "@" + StorageRefUtils.extract_storage_ref(c))
                    else:
                        provided_features_ref.append(feat)
        return ComponentUtils.clean_irrelevant_features(provided_features_ref)

    @staticmethod
    def extract_required_features_ref_from_pipe(pipe: NLUPipeline):
        """Extract provided features from component_list, which have  storage ref"""
        provided_features_ref = []
        for c in pipe.components:
            if c.loaded_from_pretrained_pipe:
                continue

            for feat in c.in_types:
                if 'embed' in feat:
                    # if StorageRefUtils.extract_storage_ref(os_components) !='':  # special edge case, some components might not have a storage ref set
                    if '@' not in feat:
                        provided_features_ref.append(feat + "@" + StorageRefUtils.extract_storage_ref(c))
                    else:
                        provided_features_ref.append(feat)

        return ComponentUtils.clean_irrelevant_features(provided_features_ref)

    @staticmethod
    def extract_sentence_embedding_conversion_candidates(pipe: NLUPipeline):
        """Extract information about embedding conversion candidates"""
        conversion_candidates_data = []
        for c in pipe.components:
            if ComponentUtils.component_has_embeddings_requirement(c) and not PipeUtils.is_trainable_pipe(pipe):
                storage_ref = StorageRefUtils.extract_storage_ref(c)
                conversion_applicable, conversion_data = PipelineCompleter.check_if_storage_ref_is_satisfied_or_get_conversion_candidate(
                    c, pipe, storage_ref)
                if conversion_applicable: conversion_candidates_data.append(conversion_data)

        return conversion_candidates_data

    @staticmethod
    def get_missing_required_features(pipe: NLUPipeline):
        """For every component_to_resolve in the pipeline"""
        provided_features_no_ref = ComponentUtils.clean_irrelevant_features(
            PipelineCompleter.extract_provided_features_refless_from_pipe(pipe))
        required_features_no_ref = ComponentUtils.clean_irrelevant_features(
            PipelineCompleter.extract_required_features_refless_from_pipe(pipe))
        provided_features_ref = ComponentUtils.clean_irrelevant_features(
            PipelineCompleter.extract_provided_features_ref_from_pipe(pipe))
        required_features_ref = ComponentUtils.clean_irrelevant_features(
            PipelineCompleter.extract_required_features_ref_from_pipe(pipe))

        is_trainable = PipeUtils.is_trainable_pipe(pipe)
        conversion_candidates = PipelineCompleter.extract_sentence_embedding_conversion_candidates(
            pipe)
        pipe.has_trainable_components = is_trainable

        required_features_ref, conversion_candidates = PipeUtils.remove_convertable_storage_refs(required_features_ref,
                                                                                                 conversion_candidates,
                                                                                                 provided_features_ref)
        provided_features_ref, required_features_ref = PipeUtils.update_converter_storage_refs_and_cols(pipe,
                                                                                                        provided_features_ref,
                                                                                                        required_features_ref)

        if is_trainable:

            trainable_index, embed_type = PipeUtils.find_trainable_embed_consumer(pipe)

            required_features_ref = []
            if embed_type is not None:
                # After resolve for a word embedding ,we must fix all NONES and set their storage refs !
                # embed consuming trainable annotators get their storage ref set here
                if len(provided_features_ref) == 0:
                    required_features_no_ref.append(embed_type)
                    if embed_type == NLP_FEATURES.CHUNK_EMBEDDINGS:
                        required_features_no_ref.append(NLP_FEATURES.WORD_EMBEDDINGS)
                if len(provided_features_ref) >= 1 and embed_type == NLP_FEATURES.CHUNK_EMBEDDINGS:
                    # This case is for when 1 Embed is preloaded and we still need to load the converter
                    if any(NLP_FEATURES.WORD_EMBEDDINGS in c for c in provided_features_ref):
                        required_features_no_ref.append(embed_type)

            if len(provided_features_ref) >= 1:
                # TODO Appraoches / Trainable models have no setStorageRef, we must set it after fitting
                pipe.components[trainable_index].storage_ref = provided_features_ref[0].split('@')[-1]

        missing_features_no_ref = set(required_features_no_ref) - set(
            provided_features_no_ref)  # - set(['text','label'])
        missing_features_ref = set(required_features_ref) - set(provided_features_ref)

        PipelineCompleter.log_resolution_status(provided_features_no_ref, required_features_no_ref,
                                                provided_features_ref, required_features_ref, is_trainable,
                                                conversion_candidates, missing_features_no_ref,
                                                missing_features_ref, )
        return missing_features_no_ref, missing_features_ref, conversion_candidates

    @staticmethod
    def add_sentence_embedding_converter(resolution_data: StorageRefConversionResolutionData) -> NluComponent:
        """ Return a Word to Sentence Embedding converter for a given Component. The input cols with match the Sentence Embedder ones
            The converter is a NLU Component Embelishement of the Spark NLP Sentence Embeddings Annotator
        """
        logger.info(f'Adding Sentence embedding conversion for Embedding Provider={resolution_data}')
        word_embedding_provider = resolution_data.component_candidate
        c = jsl_id_to_empty_component(NLP_NODE_IDS.SENTENCE_EMBEDDINGS_CONVERTER)
        storage_ref = StorageRefUtils.extract_storage_ref(word_embedding_provider)
        c.set_metadata(c.get_default_model(), 'sentence_embedding_converter',
                       NLP_NODE_IDS.SENTENCE_EMBEDDINGS_CONVERTER, 'xx', False, Licenses.open_source, storage_ref)
        c.model.setStorageRef(storage_ref)
        # set output cols
        embed_AT_out = NLP_FEATURES.SENTENCE_EMBEDDINGS + '@' + storage_ref
        c.model.setOutputCol(embed_AT_out)
        c.spark_output_column_names = [embed_AT_out]
        c.spark_input_column_names = [NLP_FEATURES.DOCUMENT, NLP_FEATURES.WORD_EMBEDDINGS + '@' + storage_ref]
        c.model.setInputCols(c.spark_input_column_names)
        return c

    @staticmethod
    def add_chunk_embedding_converter(
            resolution_data: StorageRefConversionResolutionData) -> NluComponent:
        """ Return a Word to CHUNK Embedding converter for a given Component. The input cols with match the Sentence Embedder ones
            The converter is a NLU Component Embelishement of the Spark NLP Sentence Embeddings Annotator
            The CHUNK embedder requires entities and also embeddings to generate data from. Since there could be multiple entities generators, we neeed to pass the correct one
        """
        # TODO REFACTOR
        logger.info(f'Adding Chunk embedding conversion  Provider={resolution_data} and NER Converter provider = ')
        word_embedding_provider = resolution_data.component_candidate
        entities_col = 'entities'
        embed_provider_col = word_embedding_provider.info.spark_output_column_names[0]

        c = jsl_id_to_empty_component(NLP_NODE_IDS.CHUNK_EMBEDDINGS_CONVERTER)
        c.set_metadata(c.get_default_model(),
                       NLP_NODE_IDS.CHUNK_EMBEDDINGS_CONVERTER, NLP_NODE_IDS.CHUNK_EMBEDDINGS_CONVERTER,
                       'xx',
                       False, Licenses.open_source)

        # c = nlu.embeddings_chunker.EmbeddingsChunker(annotator_class='chunk_embedder')
        storage_ref = StorageRefUtils.extract_storage_ref(word_embedding_provider)
        c.model.setStorageRef(storage_ref)
        c.info.storage_ref = storage_ref

        c.model.setInputCols(entities_col, embed_provider_col)
        c.model.setOutputCol('chunk_embeddings@' + storage_ref)
        c.info.spark_input_column_names = [entities_col, embed_provider_col]
        c.info.input_column_names = [entities_col, embed_provider_col]

        c.info.spark_output_column_names = ['chunk_embeddings@' + storage_ref]
        c.info.output_column_names = ['chunk_embeddings@' + storage_ref]
        return c

    @staticmethod
    def check_if_all_conversions_satisfied(components_for_embedding_conversion):
        """Check if all dependencies are satisfied."""
        for conversion in components_for_embedding_conversion:
            if conversion.component_candidate is not None: return False
        return True

    @staticmethod
    def check_dependencies_satisfied(missing_components, missing_storage_refs,
                                     components_for_embedding_conversion):
        """Check if all dependencies are satisfied."""
        return len(missing_components) == 0 and len(
            missing_storage_refs) == 0 and PipelineCompleter.check_if_all_conversions_satisfied(
            components_for_embedding_conversion)

    @staticmethod
    def has_licensed_components(pipe: NLUPipeline) -> bool:
        """Check if any licensed components in component_list"""
        for c in pipe.components:
            if c.license == Licenses.hc or c.license == Licenses.ocr:
                return True
        return False

    @staticmethod
    def check_same_as_last_iteration(last_missing_components, last_missing_storage_refs,
                                     last_components_for_embedding_conversion, missing_components, missing_storage_refs,
                                     components_for_embedding_conversion):
        return last_missing_components == missing_components and last_missing_storage_refs == missing_storage_refs and last_components_for_embedding_conversion == components_for_embedding_conversion

    @staticmethod
    def except_infinity_loop(reason):
        raise Exception(f"Sorry, nlu has problems building this spell, please report this issue. Problem={reason}")

    @staticmethod
    def satisfy_dependencies(pipe: NLUPipeline):
        """Feature Dependency Resolution Algorithm.
         For a given pipeline with N components, builds a DAG in reverse and satisfy each of their dependencies and child dependencies
         with a BFS approach and returns the resulting pipeline
         :param pipe: Nlu Pipe containing components for which dependencies should be satisfied
         :return: Nlu pipe with dependencies satisfied
         """
        all_features_provided = False
        is_licensed = PipelineCompleter.has_licensed_components(pipe)
        pipe.has_licensed_components = is_licensed
        is_trainable = PipeUtils.is_trainable_pipe(pipe)

        loop_count = 0
        max_loop_count = 5

        while all_features_provided == False:
            # After new components have been added, check again for the new components if requirements are met
            components_to_add = []
            missing_components, missing_storage_refs, components_for_embedding_conversion = \
                PipelineCompleter.get_missing_required_features(pipe)
            if PipelineCompleter.check_dependencies_satisfied(missing_components, missing_storage_refs,
                                                              components_for_embedding_conversion):
                # Now all features are provided
                break

            # Update last iteration variables
            last_missing_components, last_missing_storage_refs, last_components_for_embedding_conversion = missing_components, missing_storage_refs, components_for_embedding_conversion
            # Create missing base storage ref producers, i.e. embeddings
            for missing_component in missing_storage_refs:
                component = resolve_feature(missing_component, language=pipe.lang,
                                            is_licensed=is_licensed, is_trainable_pipe=is_trainable)
                if component is None:
                    continue
                if 'chunk_emb' in missing_component:
                    components_to_add.append(ComponentUtils.config_chunk_embed_converter(component))
                else:
                    components_to_add.append(component)

            # Create missing base components, storage refs are fetched in previous loop
            for missing_component in missing_components:
                components_to_add.append(
                    resolve_feature(missing_component, language=pipe.lang, is_licensed=is_licensed,
                                    is_trainable_pipe=is_trainable))

            # Create embedding converters
            for resolution_info in components_for_embedding_conversion:
                converter = None
                if 'word2chunk' == resolution_info.type:
                    converter = PipelineCompleter.add_chunk_embedding_converter(resolution_info)
                elif 'word2sentence' == resolution_info.type:
                    converter = PipelineCompleter.add_sentence_embedding_converter(resolution_info)
                if converter is not None:
                    components_to_add.append(converter)

            logger.info(f'Resolved for missing components the following NLU components : {components_to_add}')

            # Add missing components
            for new_component in components_to_add:
                if new_component:
                    logger.info(f'adding {new_component.name}')
                    pipe.add(new_component)

            # For some models we update storage ref to the resovling models storageref.
            # We need to update them so dependencies can properly be deducted as satisfied
            pipe = PipeUtils.update_bad_storage_refs(pipe)

            # Check if we are in an infinity loop
            if PipelineCompleter.check_same_as_last_iteration(last_missing_components, last_missing_storage_refs,
                                                              last_components_for_embedding_conversion,
                                                              missing_components, missing_storage_refs,
                                                              components_for_embedding_conversion):
                loop_count += 1
            else:
                loop_count = 0
            if loop_count > max_loop_count:
                PipelineCompleter.except_infinity_loop('Failure resolving feature dependencies')
        logger.info(f"{'!*'*20} ALL DEPENDENCIES SATISFIED {'!*'*20}")
        return pipe

    @staticmethod
    def check_and_fix_component_output_column_name_satisfaction(pipe: NLUPipeline):
        '''
        This function verifies that every input and output column name of a component_to_resolve is satisfied.
        If some output names are missing, it will be added by this method.
        Usually classifiers need to change their input column name, so that it matches one of the previous embeddings because they have dynamic output names
        This function performs the following steps :
        1. For each component_to_resolve we verify that all input column names are satisfied  by checking all other components output names
        2. When a input column is missing we do the following :
        2.1 Figure out the type of the missing input column. The name of the missing column should be equal to the type
        2.2 Check if there is already a component_to_resolve in the component_list, which provides this input (It should)
        2.3. When A providing component_to_resolve is found, check if storage ref matches up.
        2.4 If True for all, update provider component_to_resolve output name, or update the original component_to_resolve input name
        :return: NLU pipeline where the output and input column names of the models have been adjusted to each other
        '''
        logger.info("Fixing input and output column names")
        for component_to_check in pipe.components:
            if component_to_check.loaded_from_pretrained_pipe: continue
            input_columns = set(component_to_check.spark_input_column_names)
            # a component_to_resolve either has '' storage ref or at most 1
            logger.info(
                f'Checking for component_to_resolve {component_to_check.name} wether inputs {input_columns} is satisfied by another component_to_resolve in the component_list ', )
            for other_component in pipe.components:
                if component_to_check.name == other_component.name: continue
                output_columns = set(other_component.spark_output_column_names)
                input_columns -= output_columns  # remove provided columns

            input_columns = ComponentUtils.clean_irrelevant_features(input_columns)

            # Resolve basic mismatches, usually storage refs
            if len(input_columns) != 0 and not pipe.has_trainable_components or ComponentUtils.is_embedding_consumer(
                    component_to_check):  # fix missing column name
                # We must not only check if input satisfied, but if storage refs match! and Match Storage_refs accordingly
                logger.info(f"Fixing bad input col for C={component_to_check} untrainable component_list")
                resolved_storage_ref_cols = []
                for missing_column in input_columns:
                    for other_component in pipe.components:
                        if component_to_check.name == other_component.name: continue
                        if other_component.type == missing_column:
                            # We update the output name for the component_to_resolve which consumes our feature
                            if StorageRefUtils.has_storage_ref(
                                    other_component) and ComponentUtils.is_embedding_provider(component_to_check):
                                if ComponentUtils.are_producer_consumer_matches(component_to_check, other_component):
                                    resolved_storage_ref_cols.append(
                                        (other_component.spark_output_column_names[0], missing_column))

                            component_to_check.spark_output_column_names = [missing_column]
                            logger.info(
                                f'Resolved requirement for missing_column={missing_column} with inputs from provider={other_component.name} by col={missing_column} ')
                            other_component.model.setOutputCol(missing_column)

                for resolution, unsatisfied in resolved_storage_ref_cols:
                    component_to_check.spark_input_column_names.remove(unsatisfied)
                    component_to_check.spark_input_column_names.append(resolution)



            # Resolve training missmatches
            elif len(input_columns) != 0 and pipe.has_trainable_components:  # fix missing column name
                logger.info(f"Fixing bad input col for C={component_to_check} trainable component_list")
                # for trainable components, we change their input columns and leave other components outputs unchanged
                for missing_column in input_columns:
                    for other_component in pipe.components:
                        if component_to_check.name == other_component.name: continue
                        if other_component.type == missing_column:
                            # We update the input col name for the componenet that has missing cols
                            component_to_check.spark_input_column_names.remove(missing_column)
                            component_to_check.spark_input_column_names.append(
                                other_component.spark_output_column_names[0])
                            component_to_check.model.setInputCols(
                                component_to_check.spark_input_column_names)

                            logger.info(
                                f'Setting input col columns for component_to_resolve {component_to_check.name} to {other_component.spark_output_column_names[0]} ')

        return pipe

    @staticmethod
    def check_and_fix_nlu_pipeline(pipe: NLUPipeline):
        """Check if the NLU pipeline is ready to transform data and return it.
        If all dependencies not satisfied, returns a new NLU pipeline where dependencies and sub-dependencies are satisfied.
        Checks and resolves in the following order :
        1. Get a reference list of input features missing for the current component_list
        2. Resolve the list of missing features by adding new  Annotators to component_list
        3. Add NER Converter if required (When there is a NER model_anno_obj)
        4. Fix order and output column names
        5.

        :param pipe:
        :return:
        """
        # main entry point for Model stacking withouth pretrained pipelines
        # requirements and provided features will be lists of lists

        # 0. Clean old @AT storage ref from all columns
        # logger.info('Cleaning old AT refs')
        # pipe = PipeUtils.clean_AT_storage_refs(pipe)

        # 1. Resolve dependencies, builds a DAG in reverse and satisfies dependencies with a Breadth-First-Search approach
        # 0. Write additional metadata to the pipe pre pipe construction
        pipe = PipeUtils.add_metadata_to_pipe(pipe)

        logger.info('Satisfying dependencies')
        pipe = PipelineCompleter.satisfy_dependencies(pipe)

        # 2. Enforce naming schema <col_name>@<storage_ref> for storage_ref consumers and producers and <entity@nlu_ref> and <ner@nlu_ref> for NER and NER-Converters
        # and add NER-IOB to NER-Pretty converters for every NER model_anno_obj that is not already feeding a NER converter
        pipe = PipeUtils.enforce_AT_schema_on_pipeline_and_add_NER_converter(pipe)

        # 2.1 If Sentence Resolvers are in pipeline, all Sentence-Embeddings must feed from Chunk2Doc which stems from the entities column to resolve
        pipe = PipelineCompleter.enforce_chunk2doc_on_sentence_embeddings(pipe)

        # 3. Validate naming of output columns is correct and no error will be thrown in spark
        logger.info('Fixing column names')
        pipe = PipelineCompleter.check_and_fix_component_output_column_name_satisfaction(pipe)

        # 4. Set on every NLP Annotator the output columns
        pipe = PipeUtils.enforce_NLU_columns_to_NLP_columns(pipe)

        # 5. fix order
        logger.info('Optimizing component_list component_to_resolve order')
        pipe = PipelineCompleter.check_and_fix_component_order(pipe)

        # 6. Rename overlapping/duplicate leaf columns in the DAG
        logger.info('Renaming duplicates cols')
        pipe = PipeUtils.rename_duplicate_cols(pipe)

        # 7. enfore again because trainable pipes might mutate component_list cols
        pipe = PipeUtils.enforce_NLU_columns_to_NLP_columns(pipe)

        # 8. Write additional metadata to the pipe post pipe construction
        pipe = PipeUtils.add_metadata_to_pipe(pipe)

        # 9. For Table-QA based Pipes, we Inject a SetenceDetector for tokenizing the questions
        pipe = PipeUtils.add_sentence_detector_to_pipe_if_required(pipe)

        logger.info('Done with component_list optimizing')

        return pipe

    @staticmethod
    def check_and_fix_component_order(pipe: NLUPipeline):
        '''
        This method takes care that the order of components is the correct in such a way,that the pipeline can be iteratively processed by spark NLP.
        Column Names will not be touched. DAG Task Sort basically.
        '''
        logger.info("Starting to optimize component_to_resolve order ")

        correct_order_component_pipeline = []
        provided_features = []
        all_components_ordered = False
        unsorted_components = pipe.components
        update_last_type = False
        last_type_sorted = None
        trainable_updated = False
        pipe.components = sorted(pipe.components, key=lambda x: x.type)
        if not pipe.contains_ocr_components:
            # if OCR we must take text sorting into account. Non-OCR pipes get text provided externalyl
            provided_features.append('text')

        if pipe.contains_audio_components:
            provided_features.append(NLP_FEATURES.RAW_AUDIO)

        loop_count = 0
        max_loop_count = 10 * len(pipe.components)
        last_correct_order_component_pipeline = []
        last_provided_features = []

        while not all_components_ordered:
            if update_last_type:
                last_type_sorted = None
            else:
                update_last_type = True
            for component in unsorted_components:
                logger.info(f"Optimizing order for component_to_resolve {component.name}")
                input_columns = ComponentUtils.remove_storage_ref_from_features(
                    ComponentUtils.clean_irrelevant_features(component.spark_input_column_names.copy(), False, False))
                if last_type_sorted is None or component.type == last_type_sorted:
                    if set(input_columns).issubset(provided_features):
                        correct_order_component_pipeline.append(component)

                        # Leave pretrained component_list components untouched
                        if component.loaded_from_pretrained_pipe:
                            unsorted_components.remove(component)
                        if component in unsorted_components:
                            unsorted_components.remove(component)

                        # TODO remove storage ref from provided features ?
                        provided_features += ComponentUtils.remove_storage_ref_from_features(
                            ComponentUtils.clean_irrelevant_features(component.spark_output_column_names.copy(), False,
                                                                     False))

                        last_type_sorted = component.type
                        update_last_type = False

                        break

            if len(unsorted_components) == 0:
                all_components_ordered = True

            if not all_components_ordered and len(
                    unsorted_components) <= 2 and pipe.has_trainable_components and not trainable_updated and \
                    unsorted_components[0].trainable and 'sentence_embeddings@' in unsorted_components[
                0].spark_input_column_names:
                # special case, if trainable then we feed embed consumers on the first sentence embed provider
                # 1. Find first sent embed provider
                # 2. substitute any 'sent_embed@' consumer inputs for the provider col
                for f in provided_features:
                    if 'sentence_embeddings' in f and not trainable_updated:
                        unsorted_components[0].spark_input_column_names.remove('sentence_embeddings@')
                        if 'sentence_embeddings@' in unsorted_components[0].spark_input_column_names:
                            unsorted_components[0].spark_input_column_names.remove('sentence_embeddings@')
                        unsorted_components[0].spark_input_column_names.append(f)
                        if f not in unsorted_components[0].spark_input_column_names:  unsorted_components[
                            0].spark_input_column_names.append(f)
                        trainable_updated = True

            if not all_components_ordered and len(
                    unsorted_components) <= 2 and pipe.has_trainable_components and not trainable_updated and \
                    unsorted_components[0].trainable and 'word_embeddings@' in unsorted_components[
                0].spark_input_column_names:
                # special case, if trainable then we feed embed consumers on the first sentence embed provider
                # 1. Find first sent embed provider
                # 2. substitute any 'sent_embed@' consumer inputs for the provider col
                for f in provided_features:
                    if 'word_embeddings' in f and not trainable_updated:
                        unsorted_components[0].spark_input_column_names.remove('word_embeddings@')
                        if 'word_embeddings@' in unsorted_components[0].spark_input_column_names:  unsorted_components[
                            0].spark_input_column_names.remove(
                            'word_embeddings@')
                        unsorted_components[0].spark_input_column_names.append(f)
                        if f not in unsorted_components[0].spark_input_column_names:  unsorted_components[
                            0].spark_input_column_names.append(f)
                        trainable_updated = True

            # detect endless loop
            if last_correct_order_component_pipeline == correct_order_component_pipeline and last_provided_features == provided_features:
                loop_count += 1
            else:
                loop_count = 0
            if loop_count > max_loop_count:
                PipelineCompleter.except_infinity_loop('Failure sorting dependencies')
            last_provided_features = provided_features.copy()
            # correct_order_component_pipeline = last_correct_order_component_pipeline.copy()
            last_correct_order_component_pipeline = correct_order_component_pipeline.copy()

        pipe.components = correct_order_component_pipeline

        return pipe

    @staticmethod
    def is_storage_ref_match(embedding_consumer, embedding_provider, pipe: NLUPipeline):
        """Check for 2 components, if one provides the embeddings for the other. Makes sure that pipe_prediction_output_level matches up (chunk/sent/tok/embeds)"""
        consumer_AT_ref = ComponentUtils.extract_storage_ref_AT_notation_for_embeds(embedding_consumer, 'input')
        provider_AT_rev = ComponentUtils.extract_storage_ref_AT_notation_for_embeds(embedding_provider, 'output')
        consum_level = ComponentUtils.extract_embed_level_identity(embedding_consumer, 'input')
        provide_level = ComponentUtils.extract_embed_level_identity(embedding_provider, 'output')

        consumer_ref = StorageRefUtils.extract_storage_ref(embedding_consumer)
        provider_rev = StorageRefUtils.extract_storage_ref(embedding_provider)

        # input/output levels must match
        if consum_level != provide_level: return False

        # If storage ref dont match up, we must consult the storage_ref_2_embed mapping if it still maybe is a match, otherwise it is not.
        if consumer_ref == provider_rev: return True

        # Embed Components have have been resolved via@ have a  nlu_resolution_ref_source will match up with the consumer ref if correct embedding.
        if hasattr(embedding_provider.info, 'nlu_ref'):
            if consumer_ref == StorageRefUtils.extract_storage_ref(embedding_provider.info.nlu_ref): return True

        # If it is either  sentence_embedding_converter or chunk_embedding_converter then we gotta check what the storage ref of the inpot of those is.
        # If storage ref matches up, the providers output will match the consumer
        # if embedding_provider
        if embedding_provider.info.name in ["chunk_embedding_converter",
                                            'sentence_embedding_converter']:  # TODO FOR RESOLUTION
            nlu_ref, conv_prov_storage_ref = PipelineCompleter.get_converters_provider_info(embedding_provider,
                                                                                            pipe)

        return False

    @staticmethod
    def is_matching_level(embedding_consumer, embedding_provider):
        """Check for embedding consumer if input level matches up outputlevel of consumer
        """

    @staticmethod
    def get_converters_provider_info(embedding_provider, pipe: NLUPipeline):
        """For a component_to_resolve and a component_list, find storage_ref and """

    @staticmethod
    def enforce_chunk2doc_on_sentence_embeddings(pipe: NLUPipeline):
        """
        #If Sentence Resolvers are in pipeline, all Sentence-Embeddings must feed from Chunk2Doc which stems from
        the entities column to resolve We need to update input/output types of sentence Resolver, to the component_to_resolve
        so sorting does not get confused
        """
        if not pipe.has_licensed_components:
            return pipe
        resolvers = []
        ner_converters = []
        sentence_embeddings = []
        # Find Resolver
        for i, c in enumerate(pipe.components):
            if c.loaded_from_pretrained_pipe: continue
            # if isinstance(c.model_anno_obj, SentenceEntityResolverModel): resolvers.append(c)
            # if isinstance(c.model_anno_obj, (NerConverter, NerConverterInternal)): ner_converters.append(c)
            # if 'sentence_embeddings' == c.info.type: sentence_embeddings.append(c)
            if c.name == NLP_HC_NODE_IDS.SENTENCE_ENTITY_RESOLVER:
                resolvers.append(c)
            if c.name in [NLP_NODE_IDS.NER_CONVERTER, NLP_HC_NODE_IDS.NER_CONVERTER_INTERNAL]:
                ner_converters.append(c)
            if c.type == AnnoTypes.DOCUMENT_EMBEDDING or c.type == AnnoTypes.SENTENCE_EMBEDDING:
                sentence_embeddings.append(c)

        # No resolvers, nothing to update
        if len(resolvers) == 0:
            return pipe

        # Update Resolver
        # TODO this does not work in multi resolver scenarios reliably
        if NLP_FEATURES.DOCUMENT in sentence_embeddings[0].in_types:
            sentence_embeddings[0].in_types.remove(NLP_FEATURES.DOCUMENT)
        if NLP_FEATURES.SENTENCE in sentence_embeddings[0].in_types:
            sentence_embeddings[0].in_types.remove(NLP_FEATURES.SENTENCE)
        if NLP_FEATURES.DOCUMENT in sentence_embeddings[0].spark_input_column_names:
            sentence_embeddings[0].spark_input_column_names.remove(NLP_FEATURES.DOCUMENT)
        if NLP_FEATURES.SENTENCE in sentence_embeddings[0].spark_input_column_names:
            sentence_embeddings[0].spark_input_column_names.remove(NLP_FEATURES.SENTENCE)
        sentence_embeddings[0].in_types.append(NLP_FEATURES.DOCUMENT_FROM_CHUNK)
        sentence_embeddings[0].spark_input_column_names.append(NLP_FEATURES.DOCUMENT_FROM_CHUNK)

        # sentence_embeddings[0].info.inputs = ['chunk2doc']
        # sentence_embeddings[0].info.spark_input_column_names = ['chunk2doc']
        # sentence_embeddings[0].model_anno_obj.setInputCols('chunk2doc') # shouldb e handled by enforcing
        # chunk2doc.model_anno_obj.setOutputCol("chunk2doc")
        # chunk2doc.info.inputs = ner_converters[0].spark_output_column_names

        # TODO this will not be resolved by the resolution Algo!!
        chunk2doc = resolve_feature(NLP_FEATURES.DOCUMENT_FROM_CHUNK, 'xx')
        chunk2doc.model.setInputCols(ner_converters[0].spark_output_column_names)
        chunk2doc.spark_input_column_names = ner_converters[0].spark_output_column_names
        pipe.components.append(chunk2doc)
        # this will add a entity converter and a NER model_anno_obj if none provided
        pipe = PipelineCompleter.satisfy_dependencies(pipe)

        return pipe

    @staticmethod
    def log_resolution_status(provided_features_no_ref, required_features_no_ref, provided_features_ref,
                              required_features_ref, is_trainable, conversion_candidates, missing_features_no_ref,
                              missing_features_ref, ):
        logger.info(f"========================================================================")
        logger.info(f"Resolution Status provided_features_no_ref = {set(provided_features_no_ref)}")
        logger.info(f"Resolution Status required_features_no_ref = {set(required_features_no_ref)}")
        logger.info(f"Resolution Status provided_features_ref    = {set(provided_features_ref)}")
        logger.info(f"Resolution Status required_features_ref    = {set(required_features_ref)}")
        logger.info(f"Resolution Status is_trainable             = {is_trainable}")
        logger.info(f"Resolution Status conversion_candidates    = {conversion_candidates}")
        logger.info(f"Resolution Status missing_features_no_ref  = {set(missing_features_no_ref)}")
        logger.info(f"Resolution Status conversion_candidates    = {set(missing_features_ref)}")
        logger.info(f"========================================================================")



"""
Embellishments of primitive datatypes for logical building blocks in our universe for various domain specific concepts
Helper Classes for Type Checking and IDE Compiler Hints because NewType lib does not support typeChecking
"""


# I.e. OCR, NLP, NLP-HC,
class JslUniverse(str):
    pass

class NluRef(str):
    pass

# Lang ISO TODO define universe
class LanguageIso(str):
    pass


# Engine used for computation, i.e. Pandas, Spark, Modin
class ComputeContext(str):
    pass


# Backend used for Component
class ComponentBackend(str):
    pass


# Features provided from non JSL annotators
class ExternalFeature(str):
    pass


# external feature generator or from 3rd Party Lib
class ExternalAnno(str):
    pass


# Any JSL generated Feature from annotators
class JslFeature(str):
    pass


# Any JSL Anno reference. Unique Name defined by NLU to identify a specific class. Not using actual JVM/PY class name to make it language agnostic
class JslAnnoId(str):
    pass


# OutputLevel of a NLP annotator
class NlpLevel(str):
    pass


# Reference to Python class of an Annotator
class JslAnnoPyClass(str):
    pass


# Reference to Java class of an Annotator
class JslAnnoJavaClass(str):
    pass


# Reference to LicenseTypes
class LicenseType(str):
    pass

# Reference to Model Buckets
class ModelBucketType(str):
    pass



"""Collection of methods to substitute cols of licensed component_to_resolve results"""
import logging

logger = logging.getLogger('nlu')


def substitute_recognized_text_cols(c, cols, is_unique=True, nlu_identifier=''):
    """
    Drug Norm is always unique
    Fetched fields are:
    - entities@<storage_ref>_results
    - entities@<storage_ref>_<metadata>
        - entities@<storage_ref>_entity
        - entities@<storage_ref>_confidence
    """
    new_cols = {}
    for c in cols:
        new_cols[c] = c
    return new_cols  # TODO
    # new_base_name = 'generic_classifier' if is_unique else f'generic_classification_{nlu_identifier}'
    # for col in cols :
    #     if '_results'    in col     : new_cols[col] = new_base_name
    #     elif '_beginnings' in col     : new_cols[col] = f'{new_base_name}_begin'
    #     elif '_endings'    in col     : new_cols[col] = f'{new_base_name}_end'
    #     elif '_embeddings' in col     : continue # irrelevant  new_cols[col] = f'{new_base_name}_embedding'
    #     elif '_types'      in col     : continue # new_cols[col] = f'{new_base_name}_type'
    #     elif 'meta' in col:
    #         if   '_sentence'       in col  : new_cols[col] = f'{new_base_name}_origin_sentence'  # maps to which sentence token comes from
    #         elif   'confidence'       in col  : new_cols[col] = f'{new_base_name}_confidence'  # maps to which sentence token comes from
    #
    #     else : logger.info(f'Dropping unmatched metadata_col={col} for c={c}')
    #     # new_cols[col]= f"{new_base_name}_confidence"
    # return new_cols



from sparknlp.annotator import Chunk2Doc

class Chunk_2_Doc:
    @staticmethod
    def get_default_model():
        return Chunk2Doc() \
            .setInputCols("entities") \
            .setOutputCol("chunk2doc")






from typing import Dict

from nlu.universe.atoms import JslAnnoId, JslAnnoPyClass
from nlu.universe.feature_node_ids import OCR_NODE_IDS, NLP_NODE_IDS, NLP_HC_NODE_IDS


class AnnoClassRef:
    # Reference of every Annotator class name in OS/HC/OCR
    # Maps JslAnnoID to ClassNames in Python/Java from Spark NLP/Healthcare/ OCR
    A_O = OCR_NODE_IDS
    A_H = None  # NLP_HC_ANNO
    A_N = NLP_NODE_IDS
    HC_A_N = NLP_HC_NODE_IDS
    # Map AnnoID to PyCLass
    JSL_anno2_py_class: Dict[JslAnnoId, JslAnnoPyClass] = {

        A_N.E5_SENTENCE_EMBEDDINGS: 'E5Embeddings',
        A_N.INSTRUCTOR_SENTENCE_EMBEDDINGS: 'InstructorEmbeddings',

        A_N.WHISPER_FOR_CTC: 'WhisperForCTC',
        A_N.HUBERT_FOR_CTC: 'HubertForCTC',
        A_N.CAMEMBERT_FOR_QUESTION_ANSWERING: 'CamemBertForQuestionAnswering',
        A_N.SWIN_IMAGE_CLASSIFICATION: 'SwinForImageClassification',
        A_N.CAMEMBERT_FOR_SEQUENCE_CLASSIFICATION: 'CamemBertForSequenceClassification',
        A_N.TAPAS_FOR_QA: 'TapasForQuestionAnswering',
        A_N.TABLE_ASSEMBLER: 'TableAssembler',
        A_N.CAMEMBERT_FOR_TOKEN_CLASSIFICATION: 'CamemBertForTokenClassification',
        A_N.COREF_SPAN_BERT: 'SpanBertCorefModel',
        A_N.PARTIALLY_IMPLEMENTED: 'PartiallyIntegrated',

        A_N.BIG_TEXT_MATCHER: 'BigTextMatcher',
        A_N.CHUNK2DOC: 'Chunk2Doc',
        A_N.CHUNK_EMBEDDINGS_CONVERTER: 'ChunkEmbeddings',
        A_N.CHUNK_TOKENIZER: 'Tokenizer',
        A_N.CHUNKER: 'Chunker',
        A_N.CLASSIFIER_DL: 'ClassifierDLModel',
        A_N.CONTEXT_SPELL_CHECKER: 'ContextSpellCheckerModel',
        A_N.DATE_MATCHER: 'DateMatcher',
        A_N.UNTYPED_DEPENDENCY_PARSER: 'DependencyParserModel',
        A_N.TYPED_DEPENDENCY_PARSER: 'TypedDependencyParserModel',
        A_N.DOC2CHUNK: 'Doc2Chunk',
        A_N.DOC2VEC: 'Doc2VecModel',
        A_N.TRAIANBLE_DOC2VEC: 'Doc2VecApproach',

        A_N.MULTI_DOCUMENT_ASSEMBLER: 'MultiDocumentAssembler',
        A_N.ALBERT_FOR_QUESTION_ANSWERING: 'AlbertForQuestionAnswering',
        A_N.BERT_FOR_QUESTION_ANSWERING: 'BertForQuestionAnswering',
        A_N.DE_BERTA_FOR_QUESTION_ANSWERING: 'DeBertaForQuestionAnswering',
        A_N.DISTIL_BERT_FOR_QUESTION_ANSWERING: 'DistilBertForQuestionAnswering',
        A_N.LONGFORMER_FOR_QUESTION_ANSWERING: 'LongformerForQuestionAnswering',
        A_N.ROBERTA_FOR_QUESTION_ANSWERING: 'RoBertaForQuestionAnswering',
        A_N.XLM_ROBERTA_FOR_QUESTION_ANSWERING: 'XlmRoBertaForQuestionAnswering',

        A_N.AUDIO_ASSEMBLER: 'AudioAssembler',
        A_N.WAV2VEC_FOR_CTC: 'Wav2Vec2ForCTC',
        A_N.DOCUMENT_ASSEMBLER: 'DocumentAssembler',
        A_N.IMAGE_ASSEMBLER: 'ImageAssembler',
        A_N.DOCUMENT_NORMALIZER: 'DocumentNormalizer',
        A_N.EMBEDDINGS_FINISHER: 'EmbeddingsFinisher',
        A_N.ENTITY_RULER: 'EntityRulerModel',
        A_N.FINISHER: 'Finisher',
        A_N.GRAPH_EXTRACTION: 'GraphExtraction',
        A_N.GRAPH_FINISHER: 'GraphFinisher',
        A_N.LANGUAGE_DETECTOR_DL: 'LanguageDetectorDL',
        A_N.LEMMATIZER: 'LemmatizerModel',
        A_N.MULTI_CLASSIFIER_DL: 'MultiClassifierDLModel',
        A_N.MULTI_DATE_MATCHER: 'MultiDateMatcher',
        A_N.N_GRAMM_GENERATOR: 'NGramGenerator',
        A_N.NER_CONVERTER: 'NerConverter',
        A_N.NER_CRF: 'NerCrfModel',
        A_N.NER_DL: 'NerDLModel',
        A_N.NER_OVERWRITER: 'NerOverwriter',
        A_N.NORMALIZER: 'NormalizerModel',
        A_N.NORVIG_SPELL_CHECKER: 'NorvigSweetingModel',
        A_N.POS: 'PerceptronModel',
        A_N.RECURISVE_TOKENIZER: 'RecursiveTokenizerModel',

        A_N.REGEX_MATCHER: 'RegexMatcherModel',
        A_N.REGEX_TOKENIZER: 'RegexTokenizer',
        A_N.SENTENCE_DETECTOR: 'SentenceDetector',
        A_N.SENTENCE_DETECTOR_DL: 'SentenceDetectorDLModel',
        A_N.SENTENCE_EMBEDDINGS_CONVERTER: 'SentenceEmbeddings',
        A_N.STEMMER: 'Stemmer',
        A_N.STOP_WORDS_CLEANER: 'StopWordsCleaner',
        A_N.SYMMETRIC_DELETE_SPELLCHECKER: 'SymmetricDeleteModel',
        A_N.TEXT_MATCHER: 'TextMatcherModel',
        A_N.TOKEN2CHUNK: 'Token2Chunk',
        A_N.TOKEN_ASSEMBLER: 'TokenAssembler',
        A_N.TOKENIZER: 'TokenizerModel',
        A_N.SENTIMENT_DL: 'SentimentDLModel',
        A_N.SENTIMENT_DETECTOR: 'SentimentDetectorModel',
        A_N.VIVEKN_SENTIMENT: 'ViveknSentimentModel',
        A_N.WORD_EMBEDDINGS: 'WordEmbeddingsModel',
        A_N.WORD_SEGMENTER: 'WordSegmenterModel',
        A_N.YAKE_KEYWORD_EXTRACTION: 'YakeKeywordExtraction',
        A_N.ALBERT_EMBEDDINGS: 'AlbertEmbeddings',
        A_N.ALBERT_FOR_TOKEN_CLASSIFICATION: 'AlbertForTokenClassification',
        A_N.BERT_EMBEDDINGS: 'BertEmbeddings',
        A_N.BERT_FOR_TOKEN_CLASSIFICATION: 'BertForTokenClassification',
        A_N.BERT_SENTENCE_EMBEDDINGS: 'BertSentenceEmbeddings',
        A_N.DISTIL_BERT_EMBEDDINGS: 'DistilBertEmbeddings',
        A_N.DISTIL_BERT_FOR_SEQUENCE_CLASSIFICATION: 'DistilBertForSequenceClassification',
        A_N.DISTIL_BERT_FOR_ZERO_SHOT_CLASSIFICATION: 'DistilBertForZeroShotClassification',
        A_N.BERT_FOR_SEQUENCE_CLASSIFICATION: 'BertForSequenceClassification',
        A_N.BERT_FOR_ZERO_SHOT_CLASSIFICATION: 'BertForZeroShotClassification',
        A_N.ELMO_EMBEDDINGS: 'ElmoEmbeddings',
        A_N.LONGFORMER_EMBEDDINGS: 'LongformerEmbeddings',
        A_N.ROBERTA_FOR_ZERO_SHOT_CLASSIFICATION: 'RoBertaForZeroShotClassification',
        A_N.LONGFORMER_FOR_TOKEN_CLASSIFICATION: 'LongformerForTokenClassification',
        A_N.MARIAN_TRANSFORMER: 'MarianTransformer',
        A_N.MPNET_SENTENCE_EMBEDDINGS: 'MPNetEmbeddings',
        A_N.ROBERTA_EMBEDDINGS: 'RoBertaEmbeddings',
        A_N.ROBERTA_FOR_TOKEN_CLASSIFICATION: 'RoBertaForTokenClassification',
        A_N.ROBERTA_SENTENCE_EMBEDDINGS: 'RoBertaSentenceEmbeddings',
        A_N.T5_TRANSFORMER: 'T5Transformer',
        A_N.UNIVERSAL_SENTENCE_ENCODER: 'UniversalSentenceEncoder',
        A_N.XLM_ROBERTA_EMBEDDINGS: 'XlmRoBertaEmbeddings',
        A_N.XLM_ROBERTA_FOR_TOKEN_CLASSIFICATION: 'XlmRoBertaForTokenClassification',
        A_N.XLM_ROBERTA_SENTENCE_EMBEDDINGS: 'XlmRoBertaSentenceEmbeddings',
        A_N.XLNET_EMBEDDINGS: 'XlnetEmbeddings',
        A_N.XLNET_FOR_TOKEN_CLASSIFICATION: 'XlnetForTokenClassification',
        A_N.XLM_ROBERTA_FOR_SEQUENCE_CLASSIFICATION: 'XlmRoBertaForSequenceClassification',
        A_N.ROBERTA_FOR_SEQUENCE_CLASSIFICATION: 'RoBertaForSequenceClassification',
        A_N.LONGFORMER_FOR_SEQUENCE_CLASSIFICATION: 'LongformerForSequenceClassification',
        A_N.ALBERT_FOR_SEQUENCE_CLASSIFICATION: 'AlbertForSequenceClassification',
        A_N.XLNET_FOR_SEQUENCE_CLASSIFICATION: 'XlnetForSequenceClassification',
        A_N.GPT2: 'GPT2Transformer',
        A_N.DEBERTA_WORD_EMBEDDINGS: 'DeBertaEmbeddings',
        A_N.DEBERTA_FOR_TOKEN_CLASSIFICATION: 'DeBertaForTokenClassification',
        A_N.CAMEMBERT_EMBEDDINGS: 'CamemBertEmbeddings',
        A_N.BART_TRANSFORMER: 'BartTransformer',

        A_N.TRAINABLE_VIVEKN_SENTIMENT: 'ViveknSentimentApproach',
        A_N.TRAINABLE_SENTIMENT: 'SentimentDetector',
        A_N.TRAINABLE_SENTIMENT_DL: 'SentimentDLApproach',
        A_N.TRAINABLE_CLASSIFIER_DL: 'ClassifierDLApproach',
        A_N.TRAINABLE_MULTI_CLASSIFIER_DL: 'MultiClassifierDLApproach',
        A_N.TRAINABLE_NER_DL: 'NerDLApproach',
        A_N.TRAINABLE_NER_CRF: 'NerCrfApproach',

        A_N.TRAINABLE_POS: 'PerceptronApproach',
        A_N.TRAINABLE_DEP_PARSE_TYPED: 'TypedDependencyParserApproach',
        A_N.TRAINABLE_DEP_PARSE_UN_TYPED: 'DependencyParserApproach',
        A_N.TRAINABLE_DOC2VEC: 'Doc2VecApproach',
        A_N.TRAINABLE_ENTITY_RULER: 'EntityRulerApproach',
        A_N.TRAINABLE_LEMMATIZER: 'Lemmatizer',
        A_N.TRAINABLE_NORMALIZER: 'Normalizer',
        A_N.TRAINABLE_NORVIG_SPELL_CHECKER: 'NorvigSweetingApproach',
        A_N.TRAINABLE_RECURISVE_TOKENIZER: 'RecursiveTokenizer',
        A_N.TRAINABLE_REGEX_MATCHER: 'RegexMatcher',
        A_N.TRAINABLE_SENTENCE_DETECTOR_DL: 'SentenceDetectorDLApproach',
        A_N.TRAINABLE_WORD_EMBEDDINGS: 'WordEmbeddings',
        A_N.TRAINABLE_SYMMETRIC_DELETE_SPELLCHECKER: 'SymmetricDeleteApproach',
        A_N.TRAINABLE_TEXT_MATCHER: 'TextMatcher',
        A_N.TRAINABLE_TOKENIZER: 'Tokenizer',
        A_N.TRAINABLE_WORD_SEGMENTER: 'WordSegmenterApproach',

        A_N.DISTIL_BERT_FOR_TOKEN_CLASSIFICATION: 'DistilBertForTokenClassification',
        A_N.WORD_2_VEC: 'Word2VecModel',
        A_N.DEBERTA_FOR_SEQUENCE_CLASSIFICATION: 'DeBertaForSequenceClassification',

        A_N.BERT_SENTENCE_CHUNK_EMBEDDINGS: 'BertSentenceChunkEmbeddings',

        A_N.PARTIAL_AssertionFilterer: 'AssertionFilterer',
        A_N.PARTIAL_ChunkConverter: 'ChunkConverter',
        A_N.PARTIAL_ChunkKeyPhraseExtraction: 'ChunkKeyPhraseExtraction',
        A_N.PARTIAL_ChunkSentenceSplitter: 'ChunkSentenceSplitter',
        A_N.PARTIAL_ChunkFiltererApproach: 'ChunkFiltererApproach',
        A_N.PARTIAL_ChunkFiltererApproach: 'ChunkFiltererApproach',
        A_N.PARTIAL_ChunkFilterer: 'ChunkFilterer',
        A_N.PARTIAL_ChunkMapperApproach: 'ChunkMapperApproach',
        A_N.PARTIAL_ChunkMapperApproach: 'ChunkMapperApproach',
        A_N.PARTIAL_ChunkMapperFilterer: 'ChunkMapperFilterer',
        A_N.PARTIAL_DocumentLogRegClassifierApproach: 'DocumentLogRegClassifierApproach',
        A_N.PARTIAL_DocumentLogRegClassifierApproach: 'DocumentLogRegClassifierApproach',
        A_N.PARTIAL_DocumentLogRegClassifierModel: 'DocumentLogRegClassifierModel',
        A_N.PARTIAL_ContextualParserApproach: 'ContextualParserApproach',
        A_N.PARTIAL_ContextualParserApproach: 'ContextualParserApproach',
        A_N.PARTIAL_ReIdentification: 'ReIdentification',
        A_N.PARTIAL_NerDisambiguator: 'NerDisambiguator',
        A_N.PARTIAL_NerDisambiguatorModel: 'NerDisambiguatorModel',
        A_N.PARTIAL_AverageEmbeddings: 'AverageEmbeddings',
        A_N.PARTIAL_EntityChunkEmbeddings: 'EntityChunkEmbeddings',
        A_N.PARTIAL_ChunkMergeApproach: 'ChunkMergeApproach',
        A_N.PARTIAL_ChunkMergeApproach: 'ChunkMergeApproach',
        A_N.PARTIAL_IOBTagger: 'IOBTagger',
        A_N.PARTIAL_NerChunker: 'NerChunker',
        A_N.PARTIAL_NerConverterInternalModel: 'NerConverterInternalModel',
        A_N.PARTIAL_DateNormalizer: 'DateNormalizer',
        A_N.PARTIAL_PosologyREModel: 'PosologyREModel',
        A_N.PARTIAL_RENerChunksFilter: 'RENerChunksFilter',
        A_N.PARTIAL_ResolverMerger: 'ResolverMerger',
        A_N.PARTIAL_AnnotationMerger: 'AnnotationMerger',
        A_N.PARTIAL_Router: 'Router',
        A_N.PARTIAL_Word2VecApproach: 'Word2VecApproach',
        A_N.PARTIAL_Word2VecApproach: 'Word2VecApproach',
        A_N.PARTIAL_WordEmbeddings: 'WordEmbeddings',
        A_N.PARTIAL_EntityRulerApproach: 'EntityRulerApproach',
        A_N.PARTIAL_EntityRulerApproach: 'EntityRulerApproach',
        A_N.PARTIAL_EntityRulerModel: 'EntityRulerModel',
        A_N.PARTIAL_TextMatcherModel: 'TextMatcherModel',
        A_N.PARTIAL_BigTextMatcher: 'BigTextMatcher',
        A_N.PARTIAL_BigTextMatcherModel: 'BigTextMatcherModel',
        A_N.PARTIAL_DateMatcher: 'DateMatcher',
        A_N.PARTIAL_MultiDateMatcher: 'MultiDateMatcher',
        A_N.PARTIAL_RegexMatcher: 'RegexMatcher',
        A_N.PARTIAL_TextMatcher: 'TextMatcher',
        A_N.PARTIAL_NerApproach: 'NerApproach',
        A_N.PARTIAL_NerCrfApproach: 'NerCrfApproach',
        A_N.PARTIAL_NerCrfApproach: 'NerCrfApproach',
        A_N.PARTIAL_NerCrfApproach: 'NerCrfApproach',
        A_N.PARTIAL_NerOverwriter: 'NerOverwriter',
        A_N.PARTIAL_DependencyParserApproach: 'DependencyParserApproach',
        A_N.PARTIAL_DependencyParserApproach: 'DependencyParserApproach',
        A_N.PARTIAL_TypedDependencyParserApproach: 'TypedDependencyParserApproach',
        A_N.PARTIAL_TypedDependencyParserApproach: 'TypedDependencyParserApproach',
        A_N.PARTIAL_SentenceDetectorDLApproach: 'SentenceDetectorDLApproach',
        A_N.PARTIAL_SentenceDetectorDLApproach: 'SentenceDetectorDLApproach',
        A_N.PARTIAL_SentimentDetector: 'SentimentDetector',
        A_N.PARTIAL_ViveknSentimentApproach: 'ViveknSentimentApproach',
        A_N.PARTIAL_ViveknSentimentApproach: 'ViveknSentimentApproach',
        A_N.PARTIAL_ContextSpellCheckerApproach: 'ContextSpellCheckerApproach',
        A_N.PARTIAL_ContextSpellCheckerApproach: 'ContextSpellCheckerApproach',
        A_N.PARTIAL_NorvigSweetingApproach: 'NorvigSweetingApproach',
        A_N.PARTIAL_NorvigSweetingApproach: 'NorvigSweetingApproach',
        A_N.PARTIAL_SymmetricDeleteApproach: 'SymmetricDeleteApproach',
        A_N.PARTIAL_SymmetricDeleteApproach: 'SymmetricDeleteApproach',
        A_N.PARTIAL_ChunkTokenizer: 'ChunkTokenizer',
        A_N.PARTIAL_ChunkTokenizerModel: 'ChunkTokenizerModel',
        A_N.PARTIAL_RecursiveTokenizer: 'RecursiveTokenizer',
        A_N.PARTIAL_RecursiveTokenizerModel: 'RecursiveTokenizerModel',
        A_N.PARTIAL_Token2Chunk: 'Token2Chunk',
        A_N.PARTIAL_WordSegmenterApproach: 'WordSegmenterApproach',
        A_N.PARTIAL_WordSegmenterApproach: 'WordSegmenterApproach',
        A_N.PARTIAL_GraphExtraction: 'GraphExtraction',
        A_N.PARTIAL_Lemmatizer: 'Lemmatizer',
        A_N.PARTIAL_Normalizer: 'Normalizer',
        A_N.VIT_IMAGE_CLASSIFICATION: 'ViTForImageClassification',
        A_N.CONVNEXT_IMAGE_CLASSIFICATION: 'ConvNextImageClassifier',

    }
    JSL_anno_HC_ref_2_py_class: Dict[JslAnnoId, JslAnnoPyClass] = {
        HC_A_N.MEDICAL_QUESTION_ANSWERING: 'MedicalQuestionAnswering',
        HC_A_N.MEDICAL_TEXT_GENERATOR: 'MedicalTextGenerator',
        HC_A_N.MEDICAL_SUMMARIZER:'MedicalSummarizer',
        HC_A_N.ZERO_SHOT_NER: 'ZeroShotNerModel',
        HC_A_N.CHUNK_MAPPER_MODEL: 'ChunkMapperModel',
        HC_A_N.ASSERTION_DL: 'AssertionDLModel',
        HC_A_N.TRAINABLE_ASSERTION_DL: 'AssertionDLApproach',
        HC_A_N.ASSERTION_FILTERER: 'AssertionFilterer',
        HC_A_N.ASSERTION_LOG_REG: 'AssertionLogRegModel',
        HC_A_N.TRAINABLE_ASSERTION_LOG_REG: 'AssertionLogRegApproach',
        HC_A_N.CHUNK2TOKEN: 'Chunk2Token',
        HC_A_N.CHUNK_ENTITY_RESOLVER: '',  # DEPRECATED
        HC_A_N.TRAINABLE_CHUNK_ENTITY_RESOLVER: '',  # DEPRECATED
        HC_A_N.CHUNK_FILTERER: 'ChunkFilterer',
        HC_A_N.TRAINABLE_CHUNK_FILTERER: 'ChunkFiltererApproach',
        HC_A_N.CHUNK_KEY_PHRASE_EXTRACTION: 'ChunkKeyPhraseExtraction',
        HC_A_N.CHUNK_MERGE: 'ChunkMergeModel',
        HC_A_N.TRAINABLE_CHUNK_MERGE: 'ChunkMergeApproach',
        HC_A_N.CONTEXTUAL_PARSER: 'ContextualParserModel',
        HC_A_N.TRAIANBLE_CONTEXTUAL_PARSER: 'ContextualParserApproach',
        HC_A_N.DE_IDENTIFICATION: 'DeIdentificationModel',
        HC_A_N.TRAINABLE_DE_IDENTIFICATION: 'DeIdentification',
        HC_A_N.DOCUMENT_LOG_REG_CLASSIFIER: 'DocumentLogRegClassifierModel',
        HC_A_N.TRAINABLE_DOCUMENT_LOG_REG_CLASSIFIER: 'DocumentLogRegClassifierApproach',
        HC_A_N.DRUG_NORMALIZER: 'DrugNormalizer',
        # HC_A_N.FEATURES_ASSEMBLER  : '', # TODO spark calss>?
        HC_A_N.GENERIC_CLASSIFIER: 'GenericClassifierModel',
        HC_A_N.TRAINABLE_GENERIC_CLASSIFIER: 'GenericClassifierApproach',
        HC_A_N.IOB_TAGGER: 'IOBTagger',
        HC_A_N.MEDICAL_NER: 'MedicalNerModel',
        HC_A_N.TRAINABLE_MEDICAL_NER: 'MedicalNerApproach',
        HC_A_N.NER_CHUNKER: 'NerChunker',
        HC_A_N.NER_CONVERTER_INTERNAL: 'NerConverterInternal',
        HC_A_N.NER_DISAMBIGUATOR: 'NerDisambiguatorModel',
        HC_A_N.TRAINABLE_NER_DISAMBIGUATOR: 'NerDisambiguatorModel',
        HC_A_N.RELATION_NER_CHUNKS_FILTERER: 'NerDisambiguator',
        HC_A_N.RE_IDENTIFICATION: 'ReIdentification',
        HC_A_N.RELATION_EXTRACTION: 'RelationExtractionModel',
        HC_A_N.TRAINABLE_RELATION_EXTRACTION: 'RelationExtractionApproach',
        HC_A_N.RELATION_EXTRACTION_DL: 'RelationExtractionDLModel',
        # HC_A_N.TRAINABLE_RELATION_EXTRACTION_DL  : '',
        HC_A_N.SENTENCE_ENTITY_RESOLVER: 'SentenceEntityResolverModel',
        HC_A_N.TRAINABLE_SENTENCE_ENTITY_RESOLVER: 'SentenceEntityResolverApproach',
        HC_A_N.MEDICAL_BERT_FOR_TOKEN_CLASSIFICATION: 'MedicalBertForTokenClassifier',

        HC_A_N.MEDICAL_BERT_FOR_SEQUENCE_CLASSIFICATION: 'MedicalBertForSequenceClassification',
        HC_A_N.MEDICAL_DISTILBERT_FOR_SEQUENCE_CLASSIFICATION: 'MedicalDistilBertForSequenceClassification',
        HC_A_N.ENTITY_CHUNK_EMBEDDING: 'EntityChunkEmbeddings',
        HC_A_N.ZERO_SHOT_RELATION_EXTRACTION: 'ZeroShotRelationExtractionModel',

    }
    JSL_anno_OCR_ref_2_py_class: Dict[JslAnnoId, JslAnnoPyClass] = {
        OCR_NODE_IDS.IMAGE2TEXT: 'ImageToText',
        OCR_NODE_IDS.PDF2TEXT: 'PdfToText',
        OCR_NODE_IDS.DOC2TEXT: 'DocToText',
        OCR_NODE_IDS.BINARY2IMAGE: 'BinaryToImage',
        OCR_NODE_IDS.PDF2TEXT_TABLE: 'PdfToTextTable',
        OCR_NODE_IDS.PPT2TEXT_TABLE: 'PptToTextTable',
        OCR_NODE_IDS.DOC2TEXT_TABLE: 'DocToTextTable',
        OCR_NODE_IDS.TEXT2PDF: 'TextToPdf',
        OCR_NODE_IDS.VISUAL_DOCUMENT_CLASSIFIER: 'VisualDocumentClassifier',
        OCR_NODE_IDS.IMAGE2HOCR: 'ImageToHocr',
        OCR_NODE_IDS.IMAGE_TABLE_DETECTOR: 'ImageTableDetector',
        OCR_NODE_IDS.IMAGE_TABLE_CELL_DETECTOR: 'ImageTableCellDetector',
        OCR_NODE_IDS.IMAGE_TABLE_CELL2TEXT_TABLE: 'ImageCellsToTextTable',
        OCR_NODE_IDS.IMAGE_SPLIT_REGIONS: 'ImageSplitRegions',
    }

    @staticmethod
    def get_os_pyclass_2_anno_id_dict():
        # Flipped, maps PyClass to AnnoID
        JSL_py_class_2_anno_id: Dict[JslAnnoPyClass, JslAnnoId] = {AnnoClassRef.JSL_anno2_py_class[k]: k for k in
                                                                   AnnoClassRef.JSL_anno2_py_class}
        return JSL_py_class_2_anno_id

    @staticmethod
    def get_hc_pyclass_2_anno_id_dict():
        # Flipped, maps PyClass to AnnoID
        JSL_HC_py_class_2_anno_id: Dict[JslAnnoId, JslAnnoPyClass] = {AnnoClassRef.JSL_anno_HC_ref_2_py_class[k]: k for
                                                                      k in AnnoClassRef.JSL_anno_HC_ref_2_py_class}
        return JSL_HC_py_class_2_anno_id

    @staticmethod
    def get_ocr_pyclass_2_anno_id_dict():
        # Flipped, maps PyClass to AnnoID
        JSL_OCR_py_class_2_anno_id: Dict[JslAnnoId, JslAnnoPyClass] = {AnnoClassRef.JSL_anno_OCR_ref_2_py_class[k]: k
                                                                       for
                                                                       k in AnnoClassRef.JSL_anno_OCR_ref_2_py_class}
        return JSL_OCR_py_class_2_anno_id


# Flipped, maps PyClass to AnnoID
AnnoClassRef.JSL_OS_py_class_2_anno_id: Dict[JslAnnoPyClass, JslAnnoId] = {AnnoClassRef.JSL_anno2_py_class[k]: k for k
                                                                           in AnnoClassRef.JSL_anno2_py_class}
AnnoClassRef.JSL_HC_py_class_2_anno_id: Dict[JslAnnoId, JslAnnoPyClass] = {AnnoClassRef.JSL_anno_HC_ref_2_py_class[k]: k
                                                                           for k in
                                                                           AnnoClassRef.JSL_anno_HC_ref_2_py_class}
AnnoClassRef.JSL_OCR_py_class_2_anno_id: Dict[JslAnnoId, JslAnnoPyClass] = {
    AnnoClassRef.JSL_anno_OCR_ref_2_py_class[k]: k for k in
    AnnoClassRef.JSL_anno_OCR_ref_2_py_class}




class Deidentifier:
    @staticmethod
    def get_default_model():
        from sparknlp_jsl.annotator import DeIdentificationModel
        return DeIdentificationModel.pretrained(name = 'redl_bodypart_direction_biobert') \
            .setInputCols(["entities", "sentence", "token"]) \
            .setOutputCol("deidentified")


    @staticmethod
    def get_pretrained_model(name, language, bucket='clinical/models'):
        from sparknlp_jsl.annotator import DeIdentificationModel
        return DeIdentificationModel.pretrained(name, language,bucket) \
            .setInputCols(["entities", "sentence", "token"]) \
            .setOutputCol("deidentified")


    @staticmethod
    def get_trainable_model():
        from sparknlp_jsl.annotator import DeIdentification
        return DeIdentification\
            .setInputCols(["entities", "sentence", "token"]) \
            .setOutputCol("deidentified")



# These Spark NLP imports are not unused and will be dynamically loaded, do not remove!
from sparknlp_jsl.annotator import *
# we just use some classes here, so that intelij cleanup will not remove the imports
MedicalNerModel.name
def verify_model_licensed(class_name: str, model_path: str):
    """
    Load a licensed model_anno_obj from HDD
    """
    try:
        m = eval(class_name).load(model_path)
        return m
    except:
        print(f"Could not load Annotator class={class_name} located in {model_path}. Try updaing spark-nlp-jsl")



class IMAGE_TABLE_DETECTOR:
    @staticmethod
    def get_default_model():
        from sparkocr.transformers import ImageTableDetector
        return ImageTableDetector.pretrained("general_model_table_detection_v2", "en", "clinical/ocr") \
            .setInputCol("ocr_image") \
            .setOutputCol("region")



from sparknlp.annotator import *
from sparknlp.base import *


class ConvNextImageClassifier:
    @staticmethod
    def get_default_model():
        return ConvNextForImageClassification \
            .pretrained() \
            .setInputCols("image_assembler") \
            .setOutputCol("classes")

    @staticmethod
    def get_pretrained_model(name, language, bucket=None):
        return ConvNextForImageClassification \
            .pretrained(name, language, bucket) \
            .setInputCols("image_assembler") \
            .setOutputCol("class")








import streamlit as st
from sparknlp.annotator import *
import nlu
from nlu.pipe.utils.resolution.storage_ref_utils import StorageRefUtils
from nlu.universe.feature_node_ids import NLP_NODE_IDS


class StreamlitUtilsOS():
    classifers_OS = [ClassifierDLModel, LanguageDetectorDL, MultiClassifierDLModel, NerDLModel, NerCrfModel,
                     YakeKeywordExtraction, PerceptronModel, SentimentDLModel,
                     SentimentDetectorModel, ViveknSentimentModel, DependencyParserModel, TypedDependencyParserModel,
                     T5Transformer, MarianTransformer, NerConverter]

    @staticmethod
    def get_classifier_cols(pipe):
        classifier_cols = []
        for c in pipe.components:
            if type(c.model) in StreamlitUtilsOS.classifers_OS:
                classifier_cols += pipe.anno2final_cols[c.model]
        return classifier_cols

    @staticmethod
    def get_embed_cols(pipe):
        classifier_cols = []
        embedders = StreamlitUtilsOS.find_all_embed_components(pipe)
        for c in embedders: classifier_cols += pipe.anno2final_cols[c.model]
        return classifier_cols

    @staticmethod
    def find_embed_col(df, search_multi=False):
        """Find col that contains embed"""
        if not search_multi:
            for c in df.columns:
                if 'embed' in c: return c
        else:
            e_cols = []
            for c in df.columns:
                if 'embed' in c: e_cols.append(c)
        return e_cols

    @staticmethod
    def find_embed_component(p):
        """Find first embed  component_to_resolve in component_list"""
        for c in p.components:
            if 'embed' in c.out_types[0]: return c
        st.warning("No Embed model_anno_obj in component_list")
        return None

    @staticmethod
    def find_all_classifier_components(pipe):
        """Find ALL classifier component_to_resolve in component_list"""
        classifier_comps = []
        for c in pipe.components:
            if type(c.model) in StreamlitUtilsOS.classifers_OS: classifier_comps.append(c)
        return classifier_comps

    @staticmethod
    def find_all_embed_components(p):
        """Find ALL  embed component_to_resolve in component_list"""
        cs = []
        for c in p.components:
            if 'embed' in c.out_types[0] and 'chunk' not in c.out_types[0]: cs.append(c)
        if len(cs) == 0: st.warning("No Embed model_anno_obj in component_list")
        return cs

    @staticmethod
    def extract_name(component_or_pipe):
        name = ''
        if hasattr(component_or_pipe, 'info'):
            if hasattr(component_or_pipe, 'nlu_ref'):
                name = component_or_pipe.nlu_ref
            elif hasattr(component_or_pipe, 'storage_ref'):
                name = component_or_pipe.storage_ref
            elif hasattr(component_or_pipe, 'nlp_ref'):
                name = component_or_pipe.nlp_ref
        elif hasattr(component_or_pipe, 'nlu_ref'):
            name = component_or_pipe.nlu_ref
        return name

    @staticmethod
    def find_ner_model(p):
        """Find NER component_to_resolve in component_list"""
        from sparknlp.annotator import NerDLModel, NerCrfModel
        for c in p.components:
            if isinstance(c.model, (NerDLModel, NerCrfModel)): return c.model
        st.warning("No NER model_anno_obj in component_list")
        return None

    @staticmethod
    def get_NER_tags_in_pipe(p):
        """Get NER tags in component_list, used for showing visualizable tags"""
        n = StreamlitUtilsOS.find_ner_model(p)
        if n is None: return []
        classes_predicted_by_ner_model = n.getClasses()
        split_iob_tags = lambda s: s.split('-')[1] if '-' in s else ''
        classes_predicted_by_ner_model = list(map(split_iob_tags, classes_predicted_by_ner_model))
        while '' in classes_predicted_by_ner_model: classes_predicted_by_ner_model.remove('')
        classes_predicted_by_ner_model = list(set(classes_predicted_by_ner_model))
        return classes_predicted_by_ner_model

    @staticmethod
    def get_manifold_algo(algo, dim, n_jobs=None):
        from sklearn.manifold import TSNE, Isomap, LocallyLinearEmbedding, MDS, SpectralEmbedding
        from sklearn.decomposition import TruncatedSVD, DictionaryLearning, FactorAnalysis, FastICA, KernelPCA, PCA, \
            LatentDirichletAllocation
        # manifold
        if algo == 'TSNE': return TSNE(n_components=dim, n_jobs=n_jobs)
        if algo == 'ISOMAP': return Isomap(n_components=dim, n_jobs=n_jobs)
        if algo == 'LLE': return LocallyLinearEmbedding(n_components=dim, n_jobs=n_jobs)
        if algo == 'Spectral Embedding': return SpectralEmbedding(n_components=dim, n_jobs=n_jobs)
        if algo == 'MDS': return MDS(n_components=dim, n_jobs=n_jobs)
        # Matrix Decomposition
        if algo == 'PCA': return PCA(n_components=dim)  # No hyper
        if algo == 'SVD aka LSA': return TruncatedSVD(n_components=dim)  # No hyper
        if algo == 'DictionaryLearning': return DictionaryLearning(n_components=dim, n_jobs=n_jobs)
        if algo == 'FactorAnalysis': return FactorAnalysis(n_components=dim)  # no hyper
        if algo == 'FastICA': return FastICA(n_components=dim)  # no hyper
        if algo == 'KernelPCA': return KernelPCA(n_components=dim, n_jobs=n_jobs)
        # not applicable because negative values, todo we could just take absolute values of all embeds..
        if algo == 'LatentDirichletAllocation': return LatentDirichletAllocation(n_components=dim)
        # if algo =='NMF': return NMF(n_components=dim)

    @staticmethod
    @st.cache(allow_output_mutation=True, hash_funcs={"_thread.RLock": lambda _: None})
    def get_pipe(model='ner'):
        return nlu.load(model)

    @staticmethod
    def merge_token_classifiers_with_embed_pipe(embed_pipe, token_pipe):
        """Merge token feature generators into embed component_list. i.e. Pos/Dep_depdency/Untyped_dep if not already present in component_list"""
        for c in token_pipe.components:

            if c.name == NLP_NODE_IDS.POS:
                for emb_c in embed_pipe.components:
                    if emb_c.name == NLP_NODE_IDS.POS:
                        embed_pipe.is_fitted = False
                        embed_pipe.fit()
                        return embed_pipe
                    # only merge if pos not already in component_list
                embed_pipe.components.append(c)
        embed_pipe.is_fitted = False
        embed_pipe.fit()
        return embed_pipe

    @staticmethod
    def extract_all_sentence_storage_refs_or_nlu_refs(e_coms):
        """extract either NLU_ref or storage_ref as fallback for a list of embedding components"""
        loaded_storage_refs = []
        loaded_embed_nlu_refs = []
        for c in e_coms:
            if not hasattr(c, 'nlu_ref'): continue
            r = c.nlu_ref
            if 'en.' not in r and 'embed_sentence.' not in r and 'ner' not in r:
                loaded_embed_nlu_refs.append('en.embed_sentence.' + r)
            elif 'en.' in r and 'embed_sentence.' not in r and 'ner' not in r:
                r = r.split('en.')[0]
                loaded_embed_nlu_refs.append('en.embed_sentence.' + r)
            else:
                loaded_embed_nlu_refs.append(StorageRefUtils.extract_storage_ref(c))
            loaded_storage_refs.append(StorageRefUtils.extract_storage_ref(c))
        return loaded_embed_nlu_refs, loaded_storage_refs



import logging

from sparknlp.annotator import *

import nlu
from nlu import Licenses
from nlu.pipe.extractors.extractor_base_data_classes import FinisherExtractorConfig
from nlu.pipe.nlu_component import NluComponent
from nlu.pipe.pipeline import NLUPipeline
from nlu.pipe.utils.resolution.storage_ref_utils import StorageRefUtils
from nlu.universe.atoms import JslAnnoId
from nlu.universe.component_universes import ComponentUniverse, jsl_id_to_empty_component
from nlu.universe.feature_node_ids import NLP_NODE_IDS, NLP_HC_NODE_IDS, OCR_NODE_IDS
from nlu.universe.feature_universes import NLP_FEATURES
from nlu.universe.logic_universes import NLP_LEVELS, AnnoTypes

logger = logging.getLogger('nlu')
from nlu.pipe.utils.component_utils import ComponentUtils
from typing import List, Union, Dict
from nlu.universe.annotator_class_universe import AnnoClassRef
from nlu.utils.environment.env_utils import is_running_in_databricks
import os
import glob
import json


class PipeUtils:
    """Pipe Level logic operations and utils"""

    @staticmethod
    def update_bad_storage_refs(pipe: NLUPipeline):
        """
        Some models have bad storage refs. The list of these bad models is defined by nlu.spellbook.Spellbook.bad_storage_refs.
        The correct storage ref is given by the resolving models storage ref defined by nlu.Spellbook.licensed_storage_ref_2_nlu_ref[pipe.lang][storage_ref].
        Once the resolving model_anno_obj is loaded in the pipe, this method will take its storage ref and write it to the bad_storage_ref model_anno_obj defined by nlu.spellbook.Spellbook.bad_storage_refs.
        If storage ref is already updated, this method will leave the pipe unchanged.
        We only check for healthcare storage refs
        :param pipe: Pipe to update bad storage refs on
        :return: Pipe where each component_to_resolve has storage ref updated, if it was not already updated

        """
        for bad_storage_ref_component in pipe.components:
            if not bad_storage_ref_component.loaded_from_pretrained_pipe and bad_storage_ref_component.has_storage_ref:
                storage_ref = StorageRefUtils.extract_storage_ref(bad_storage_ref_component)
                # After updating the storage ref it will not be in the licensed_storage_ref_2_nlu_ref mapping anymore,
                if storage_ref in nlu.spellbook.Spellbook.bad_storage_refs:
                    # since its a bad storage ref, we can resolve its storage ref by checking licensed_storage_ref_2_nlu_ref
                    if pipe.lang in nlu.Spellbook.licensed_storage_ref_2_nlu_ref.keys():
                        if storage_ref in nlu.Spellbook.licensed_storage_ref_2_nlu_ref[pipe.lang].keys():
                            storage_ref_resolver_nlu_ref = nlu.Spellbook.licensed_storage_ref_2_nlu_ref[pipe.lang][
                                storage_ref]
                            for storage_resolver in pipe.components:
                                if storage_resolver.nlu_ref == storage_ref_resolver_nlu_ref:
                                    # Update the storage ref of the bad_component to the storage ref of the resolving model_anno_obj according to licensed_storage_ref_2_nlu_ref
                                    resolving_storage_ref = StorageRefUtils.extract_storage_ref(storage_resolver)
                                    bad_storage_ref_component.model.set(bad_storage_ref_component.model.storageRef,
                                                                        resolving_storage_ref)
        return pipe

    @staticmethod
    def update_relation_extractor_models_storage_ref(pipe: NLUPipeline):
        # if provided, because the sometimes have unresolvable storage refs
        # we can find the actual storage ref only after its mapped is sresolved to an model_anno_obj defined by an nlp ref
        # If RelationExtractor is not loaded from a pretrained pipe we update its storage ref to the resolving models storage ref

        for relation_extractor_component in pipe.components:
            if relation_extractor_component.jsl_anno_class_id == NLP_HC_NODE_IDS.RELATION_EXTRACTION and not relation_extractor_component.loaded_from_pretrained_pipe:
                storage_ref = StorageRefUtils.extract_storage_ref(relation_extractor_component)
                # After updating the storage ref it will not be in the licensed_storage_ref_2_nlu_ref mapping anymore,
                # so we have to check here if it exists in the mapping before accessing it
                if pipe.lang in nlu.Spellbook.licensed_storage_ref_2_nlu_ref.keys():
                    if storage_ref in nlu.Spellbook.licensed_storage_ref_2_nlu_ref[pipe.lang].keys():
                        # We need to find  a component_to_resolve in the pipeline which has this storage ref
                        storage_ref_resolver_nlu_ref = nlu.Spellbook.licensed_storage_ref_2_nlu_ref[pipe.lang][
                            storage_ref]
                        for storage_resolver in pipe.components:
                            if storage_resolver.nlu_ref == storage_ref_resolver_nlu_ref:
                                # Update the storage ref of the RL-Extractor to the storage ref of the resolving model_anno_obj according to licensed_storage_ref_2_nlu_ref
                                resolving_storage_ref = StorageRefUtils.extract_storage_ref(storage_resolver)
                                relation_extractor_component.model.set(relation_extractor_component.model.storageRef,
                                                                       resolving_storage_ref)
        return pipe

    @staticmethod
    def get_json_data_for_pipe_model_at_stage_number(pipe_path, stage_number_as_string):
        """Gets the json metadata from a model_anno_obj for a given base path at a specific stage index"""
        c_metadata_path = f'{pipe_path}/stages/{stage_number_as_string}_*/metadata/part-00000'
        c_metadata_path = glob.glob(f'{c_metadata_path}*')[0]
        with open(c_metadata_path, "r", encoding="utf8") as f:
            data = json.load(f)
        return data

    @staticmethod
    def get_json_data_for_pipe_model_at_stage_number_on_databricks(nlp_ref, lang, digit_str):
        """Gets the json metadata from a model_anno_obj for a given base path at a specific stage index on databricks"""
        import sparknlp
        spark = sparknlp.start()
        pipe_df = spark.read.json(
            f'dbfs:/root/cache_pretrained/{nlp_ref}_{lang}*/stages/{digit_str}_*/metadata/part-00000')
        data = pipe_df.toPandas().to_dict()
        data = {k: v[0] for k, v in data.items()}
        if 'inputCols' in data['paramMap'].keys():
            data['paramMap']['inputCols'] = data['paramMap']['inputCols'].tolist()
        data
        return data


    @staticmethod
    def set_column_values_on_components_from_pretrained_pipe(component_list: List[NluComponent]):
        """Set output/input cols on Nlu Components loaded from a pipeline
        """

        for c in component_list:
            if hasattr(c.model,'inputCols') :
                inp = c.model.getInputCols()
            else:
                inp = c.model.getInputCol()

            if hasattr(c.model,'outputCols') :
                out = c.model.getOutputCols()
            else:
                out = c.model.getOutputCol()
            c.spark_input_column_names = inp if isinstance(inp, List) else [inp]
            c.spark_output_column_names = out if isinstance(out, List) else [out]

        return component_list

    @staticmethod
    def set_column_values_on_components_from_pretrained_pipe_from_disk_data(component_list: List[NluComponent], nlp_ref, lang, path):
        """Since output/input cols cannot be fetched from Annotators via get input/output col reliably, we must check
        annotator data to find them Expects a list of NLU Component objects which all stem from the same pipeline
        defined by nlp_ref
        """

        if path:
            pipe_path = path
        else:
            pipe_path = os.path.expanduser('~') + '/cache_pretrained/' + f'{nlp_ref}_{lang}'
            # We do not need to check for Spark Version, since cols should match across versions
            pipe_path = glob.glob(f'{pipe_path}*')
            if len(pipe_path) == 0:
                # try databricks env path
                if is_running_in_databricks():
                    pipe_path = [f'dbfs:/root/cache_pretrained/{nlp_ref}_{lang}']
                else:
                    raise FileNotFoundError(f"Could not find downloaded Pipeline at path={pipe_path}")
            pipe_path = pipe_path[0]
            if not os.path.exists(pipe_path) and not is_running_in_databricks():
                raise FileNotFoundError(f"Could not find downloaded Pipeline at path={pipe_path}")

        # Find HDD location of component_list and read out input/output cols
        digits_num = len(str(len(component_list)))
        digit_str = '0' * digits_num
        digit_cur = 0

        for c in component_list:
            model_name = c.model.uid.split('_')[0]
            if is_running_in_databricks():
                data = PipeUtils.get_json_data_for_pipe_model_at_stage_number_on_databricks(nlp_ref, lang, digit_str)
            else:
                data = PipeUtils.get_json_data_for_pipe_model_at_stage_number(pipe_path, digit_str)

            if 'inputCols' in data['paramMap'].keys():
                inp = data['paramMap']['inputCols']
                c.model.setInputCols(inp)
            else:
                inp = data['paramMap']['inputCol']
                c.model.setInputCol(inp)

            if 'outputCol' in data['paramMap'].keys():
                out = data['paramMap']['outputCol']
            else:
                # Sometimes paramMap is missing outputCol, so we have to use this hack
                if model_name == 'DocumentAssembler':
                    out = 'document'
                elif model_name == 'Finisher':
                    out = 'finished'

                else:
                    out = c.model.uid.split('_')[0] + '_out'

            c.spark_input_column_names = inp if isinstance(inp, List) else [inp]
            c.spark_output_column_names = [out]

            if model_name != 'Finisher':
                # finisher dynamically generates cols from input cols
                c.model.setOutputCol(out) if hasattr(c.model, 'setOutputCol') else c.model.setOutputCols(out)
            digit_cur += 1
            digit_str = str(digit_cur)
            while len(digit_str) < digits_num:
                digit_str = '0' + digit_str
        return component_list

    @staticmethod
    def is_trainable_pipe(pipe: NLUPipeline):
        """Check if component_list is trainable"""
        for c in pipe.components:
            if c.trainable: return True
        return False

    @staticmethod
    def enforece_AT_embedding_provider_output_col_name_schema_for_list_of_components(pipe_list: List[NluComponent]):
        """For every embedding provider, enforce that their output col is named <pipe_prediction_output_level>@storage_ref for
        output_levels word,chunk,sentence aka document , TODO update the classifier models swell i.e.
        word_embed@elmo or sentence_embed@elmo etc. """
        for c in pipe_list:
            if ComponentUtils.is_embedding_provider(c):
                level_AT_ref = ComponentUtils.extract_storage_ref_AT_notation_for_embeds(c, 'output')
                c.out_types = [level_AT_ref]
                c.info.spark_output_column_names = [level_AT_ref]
                c.model.setOutputCol(level_AT_ref[0])
        return pipe_list

    @staticmethod
    def enforce_AT_schema_on_pipeline_and_add_NER_converter(pipe: NLUPipeline):
        """Enforces the AT naming schema on all column names and add missing NER converters"""
        return PipeUtils.enforce_AT_schema_on_NER_processors_and_add_missing_NER_converters(
            PipeUtils.enforce_AT_schema_on_embedding_processors(pipe))

    @staticmethod
    def enforce_AT_schema_on_NER_processors_and_add_missing_NER_converters(pipe: NLUPipeline):
        """For every NER provider and consumer, enforce that their output col is named <pipe_prediction_output_level>@storage_ref for
        output_levels word,chunk,sentence aka document , i.e. word_embed@elmo or sentence_embed@elmo etc. We also
        add NER converters for every NER model_anno_obj that no Converter converting its inputs In addition, returns the
        pipeline with missing NER converters added, for every NER model_anno_obj. The converters transform the IOB schema in a
        merged and more usable form for downstream tasks 1. Find a NER model_anno_obj in component_list 2. Find a NER
        converter feeding from it, if there is None, create one. 3. Generate name with Identifier
        <ner-iob>@<nlu_ref_identifier>  and <entities>@<nlu_ref_identifier> 3.1 Update NER Models    output to
        <ner-iob>@<nlu_ref_identifier> 3.2 Update NER Converter input  to <ner-iob>@<nlu_ref_identifier> 3.3 Update
        NER Converter output to <entities>@<nlu_ref_identifier> 4. Update every Component that feeds from the NER
        converter (i.e. Resolver etc.)

        includes TOKEN-CLASSIFIER-TRANSFORMER models which usually output NER format
        """
        new_converters = []

        for c in pipe.components:
            if c.loaded_from_pretrained_pipe:
                # Leave pretrained component_list models untouched
                new_converters.append(c)
                continue

            # TRANSFORMER_TOKEN_CLASSIFIER might be a NER provider. Regardless, No ner-Conversion will be performed
            # because it will not return NER IOB
            if ComponentUtils.is_NER_provider(c):
                if c.type == AnnoTypes.TRANSFORMER_TOKEN_CLASSIFIER and not ComponentUtils.is_NER_IOB_token_classifier(
                        c):
                    continue
                output_NER_col = ComponentUtils.extract_NER_col(c, 'output')
                converter_to_update = None
                for other_c in pipe.components:
                    if output_NER_col in other_c.spark_input_column_names and ComponentUtils.is_NER_converter(other_c):
                        converter_to_update = other_c

                ner_identifier = ComponentUtils.get_nlu_ref_identifier(c)
                if converter_to_update is None:
                    if c.license == Licenses.hc:
                        converter_to_update = jsl_id_to_empty_component(NLP_HC_NODE_IDS.NER_CONVERTER_INTERNAL)
                        converter_to_update.set_metadata(converter_to_update.get_default_model(),
                                                         NLP_HC_NODE_IDS.NER_CONVERTER_INTERNAL,
                                                         NLP_HC_NODE_IDS.NER_CONVERTER_INTERNAL,
                                                         'xx', False, Licenses.hc)
                    else:
                        converter_to_update = jsl_id_to_empty_component(NLP_NODE_IDS.NER_CONVERTER)
                        converter_to_update.set_metadata(converter_to_update.get_default_model(),
                                                         NLP_NODE_IDS.NER_CONVERTER, NLP_NODE_IDS.NER_CONVERTER,
                                                         'xx', False, Licenses.open_source)

                    new_converters.append(converter_to_update)
                converter_to_update.nlu_ref = f'ner_converter.{c.nlu_ref}'

                # 3. generate new col names
                new_NER_AT_ref = output_NER_col
                if '@' not in output_NER_col: new_NER_AT_ref = output_NER_col + '@' + ner_identifier
                new_NER_converter_AT_ref = 'entities' + '@' + ner_identifier

                # 3.1 upate NER model_anno_obj outputs
                c.spark_output_column_names = [new_NER_AT_ref]
                c.model.setOutputCol(new_NER_AT_ref)

                # 3.2 update converter inputs
                old_ner_input_col = ComponentUtils.extract_NER_converter_col(converter_to_update, 'input')
                if old_ner_input_col in converter_to_update.spark_input_column_names:
                    converter_to_update.spark_input_column_names.remove(old_ner_input_col)
                else:
                    converter_to_update.spark_input_column_names.pop()

                # if old_ner_input_col in converter_to_update.spark_input_column_names:
                #     converter_to_update.spark_input_column_names.remove(old_ner_input_col)
                # else:
                #     converter_to_update.spark_input_column_names.pop()
                converter_to_update.spark_input_column_names.append(new_NER_AT_ref)
                converter_to_update.model.setInputCols(converter_to_update.spark_input_column_names)

                # 3.3 update converter outputs
                converter_to_update.spark_output_column_names = [new_NER_converter_AT_ref]
                converter_to_update.model.setOutputCol(new_NER_converter_AT_ref)

                ## todo improve, this causes the first ner producer to feed to all ner-cosnuners. All other ner-producers will be ignored by ner-consumers,w ithouth special syntax or manual configs --> Chunk merger
                ##4. Update all NER consumers input columns, i.e. Resolver, Relation, etc..
                for conversion_consumer in pipe.components:
                    if NLP_FEATURES.NAMED_ENTITY_CONVERTED in conversion_consumer.in_types:
                        conversion_consumer.spark_input_column_names.remove(NLP_FEATURES.NAMED_ENTITY_CONVERTED)
                        conversion_consumer.spark_input_column_names.append(new_NER_converter_AT_ref)

        # Add new converters to component_list
        for conv in new_converters:
            if conv.license == Licenses.hc:
                pipe.add(conv,
                         name_to_add=f'chunk_converter_licensed@{conv.spark_output_column_names[0].split("@")[0]}')
            else:
                pipe.add(conv, name_to_add=f'chunk_converter@{conv.spark_output_column_names[0].split("@")[0]}')
        return pipe

    @staticmethod
    def enforce_AT_schema_on_embedding_processors(pipe: NLUPipeline):
        """For every embedding provider and consumer, enforce that their output col is named
        <pipe_prediction_output_level>@storage_ref for output_levels word,chunk,sentence aka document , i.e. word_embed@elmo or
        sentence_embed@elmo etc. """
        for c in pipe.components:
            # Leave pretrained component_list models untouched
            if c.loaded_from_pretrained_pipe: continue
            if ComponentUtils.is_embedding_provider(c):
                if '@' not in c.spark_output_column_names[0]:
                    new_embed_AT_ref = ComponentUtils.extract_storage_ref_AT_notation_for_embeds(c, 'output')
                    c.spark_output_column_names = [new_embed_AT_ref]
                    c.model.setOutputCol(new_embed_AT_ref)
            if ComponentUtils.is_embedding_consumer(c):
                input_embed_col = ComponentUtils.extract_embed_col(c)
                if '@' not in input_embed_col:
                    # TODO set storage ref for traianble model_anno_obj?
                    new_embed_AT_ref = ComponentUtils.extract_storage_ref_AT_notation_for_embeds(c, 'input')
                    c.spark_input_column_names.remove(input_embed_col)
                    c.spark_input_column_names.append(new_embed_AT_ref)
                    c.model.setInputCols(c.spark_input_column_names)

        return pipe

    @staticmethod
    def enforce_NLU_columns_to_NLP_columns(pipe: NLUPipeline):
        """for every component_to_resolve, set its inputs and outputs to the ones configured on the NLU component_to_resolve."""
        # These anno have no standardized setInputCol or it should not be configured
        blacklisted = [NLP_NODE_IDS.DOCUMENT_ASSEMBLER]
        for c in pipe.components:
            if c.name == OCR_NODE_IDS.VISUAL_DOCUMENT_CLASSIFIER:
                c.model.setLabelCol(c.spark_output_column_names[0])
                c.model.setConfidenceCol(c.spark_output_column_names[1])
                continue
            if c.loaded_from_pretrained_pipe:
                continue
            if c.name in blacklisted:
                continue

            if hasattr(c.model, 'setOutputCol'):
                c.model.setOutputCol(c.spark_output_column_names[0])
            else:
                c.model.setOutputCols(c.spark_output_column_names)

            if hasattr(c.model, 'setInputCols'):
                c.model.setInputCols(c.spark_input_column_names)
            else:
                # Some OCR Annotators only have one input and thus only setInputCol method but not setInputCols
                c.model.setInputCol(c.spark_input_column_names[0])

        return pipe

    @staticmethod
    def is_converter_component_resolution_reference(reference: str) -> bool:
        if 'chunk_emb' in reference:
            return True

    @staticmethod
    def configure_component_output_levels_to_sentence(pipe: NLUPipeline):
        '''
        Configure component_list components to output level document. Substitute every occurrence of <document> to
        <sentence> for every component_to_resolve that feeds from <document
        :param pipe: component_list to be configured
        :return: configured component_list
        '''
        logger.info('Configuring components to sentence level')
        for c in pipe.components:
            # update in/out spark cols
            if c.loaded_from_pretrained_pipe:
                continue
            if NLP_FEATURES.DOCUMENT in c.spark_input_column_names and NLP_FEATURES.SENTENCE not in c.spark_input_column_names and NLP_FEATURES.SENTENCE not in c.spark_output_column_names:
                logger.info(f"Configuring C={c.name}  of Type={type(c.model)} to Sentence Level")
                c.spark_input_column_names.remove(NLP_FEATURES.DOCUMENT)
                c.spark_input_column_names.append(NLP_FEATURES.SENTENCE)
                c.model.setInputCols(c.spark_input_column_names)
                if 'input_dependent' in c.output_level:
                    c.output_level = NLP_LEVELS.SENTENCE
            # update in/out col types
            if NLP_FEATURES.DOCUMENT in c.in_types and NLP_FEATURES.SENTENCE not in c.in_types and NLP_FEATURES.SENTENCE not in c.out_types:
                c.in_types.remove(NLP_FEATURES.DOCUMENT)
                c.in_types.append(NLP_FEATURES.SENTENCE)
        return pipe.components

    @staticmethod
    def configure_component_output_levels_to_document(pipe: NLUPipeline):
        '''
        Configure component_list components to output level document. Substitute every occurence of <sentence> to <document> for every component_to_resolve that feeds from <sentence>
        :param pipe: component_list to be configured
        :return: configured component_list coonents only
        '''
        logger.info('Configuring components to document level')
        for c in pipe.components:
            if c.loaded_from_pretrained_pipe:
                continue
            # Update in/out spark cols
            if NLP_FEATURES.SENTENCE in c.spark_input_column_names and NLP_FEATURES.DOCUMENT not in c.spark_input_column_names and NLP_FEATURES.DOCUMENT not in c.spark_output_column_names:
                logger.info(f"Configuring C={c.name} to document output level")
                c.spark_input_column_names.remove(NLP_FEATURES.SENTENCE)
                c.spark_input_column_names.append(NLP_FEATURES.DOCUMENT)
                c.model.setInputCols(c.spark_input_column_names)
                if 'input_dependent' in c.output_level:
                    c.output_level = NLP_LEVELS.DOCUMENT
            # Update in/out col types
            if NLP_FEATURES.SENTENCE in c.in_types and NLP_FEATURES.DOCUMENT not in c.in_types and NLP_FEATURES.DOCUMENT not in c.out_types:
                c.in_types.remove(NLP_FEATURES.SENTENCE)
                c.in_types.append(NLP_FEATURES.DOCUMENT)
        return pipe.components

    @staticmethod
    def has_component_with_id(pipe: NLUPipeline, ids: Union[JslAnnoId, List[JslAnnoId]]):
        """Check for NLUPipeline if it contains component with id """
        ids = ids if isinstance(ids, list) else [ids]
        for c in pipe.components:
            if c.name in ids:
                return True
        return False

    @staticmethod
    def has_sentence_detector(pipe: NLUPipeline):
        """Check for NLUPipeline if it contains sentence detector"""
        for c in pipe.components:
            if isinstance(c.model, (SentenceDetectorDLModel, SentenceDetector, SentenceDetectorDLApproach)): return True
        return False

    @staticmethod
    def has_document_assembler(pipe: NLUPipeline):
        return PipeUtils.has_component_with_id(pipe, NLP_NODE_IDS.DOCUMENT_ASSEMBLER)

    @staticmethod
    def has_table_extractor(pipe: NLUPipeline):
        """Check for NLUPipieline if it contains any table extracting OCR component"""
        return PipeUtils.has_component_with_id(pipe, [OCR_NODE_IDS.PDF2TEXT_TABLE,
                                                      OCR_NODE_IDS.PPT2TEXT_TABLE,
                                                      OCR_NODE_IDS.DOC2TEXT_TABLE,
                                                      OCR_NODE_IDS.IMAGE_TABLE_DETECTOR,
                                                      ])

    @staticmethod
    def get_component_idx_by_id(pipe: NLUPipeline, node_id: JslAnnoId):
        """Find first occurrence of component in pipe by ID and returns index """
        for i, c in enumerate(pipe.components):
            if c.name == node_id:
                return i
        raise Exception(f'Could not find component {node_id} in pipe {pipe}')

    @staticmethod
    def add_tokenizer_to_pipe_if_missing(pipe: NLUPipeline):
        """add tokenizer to pipe if it is missing
        :param pipe:  pipe
        :return: Pipe with tokenizer if missing
        """

        if PipeUtils.has_component_with_id(pipe, [NLP_NODE_IDS.TOKENIZER, NLP_NODE_IDS.TOKEN_ASSEMBLER,
                                                  NLP_NODE_IDS.REGEX_TOKENIZER,
                                                  NLP_NODE_IDS.RECURISVE_TOKENIZER, NLP_NODE_IDS.WORD_SEGMENTER]):
            return pipe

        from nlu.pipe.component_resolution import resolve_feature
        tokenizer = resolve_feature(NLP_FEATURES.TOKEN)
        tokenizer.spark_input_column_names = [pipe.component_output_level]
        tokenizer.spark_output_column_names = [NLP_FEATURES.TOKEN]
        tokenizer.model.setInputCols(pipe.component_output_level)
        tokenizer.model.setOutputCol(NLP_FEATURES.TOKEN)

        # Find the document/sentence component and add tokenizer right after that
        for i, c in enumerate(pipe.components):
            if pipe.component_output_level in c.spark_output_column_names:
                pipe.components.insert(i + 1, tokenizer)

        return pipe

    @staticmethod
    def configure_component_output_levels(pipe: NLUPipeline, new_output_level=''):
        '''
        This method configures sentenceEmbeddings and Classifier components to output at a specific level.
        Generally this substitutes all `sentence` columns to `document` and vice versa.
        Adds SentenceDetector to pipeline if none exists
        This method is called the first time .predict() is called and every time the pipe_prediction_output_level changed
        If pipe_prediction_output_level == Document, then sentence embeddings will be fed on Document col and
            classifiers receive doc_embeds/doc_raw column,
            depending on if the classifier works with or without embeddings
        If pipe_prediction_output_level == sentence, then sentence embeddings will be fed on sentence col and
            classifiers receive sentence_embeds/sentence_raw column,
            depending on if the classifier works with or without embeddings.
            If sentence detector is missing, one will be added.
        :param pipe: NLU pipeline
        :param new_output_level: The new output level to apply, either sentence or document
        :return: Nlu pipeline, with all components output levels configured to new_output_level
        '''
        if not PipeUtils.has_document_assembler(pipe):
            # When loaded from OCR, we might not have a documentAssembler in pipe
            pipe.is_fitted = False
            document_assembler = ComponentUniverse.components[NLP_NODE_IDS.DOCUMENT_ASSEMBLER]()
            document_assembler.set_metadata(document_assembler.get_default_model(), 'document_assembler',
                                            'document_assembler', 'xx', False, Licenses.open_source)
            pipe.components.insert(0, document_assembler)

        if new_output_level == 'sentence':
            if not PipeUtils.has_sentence_detector(pipe):
                logger.info("Adding missing Sentence Detector")
                pipe.is_fitted = False
                sentence_detector = ComponentUniverse.components[NLP_NODE_IDS.SENTENCE_DETECTOR_DL]()
                sentence_detector.set_metadata(sentence_detector.get_default_model(), 'detect_sentence',
                                               'sentence_detector_dl', 'en', False, Licenses.open_source)
                insert_idx = PipeUtils.get_component_idx_by_id(pipe, NLP_NODE_IDS.DOCUMENT_ASSEMBLER)
                # insert After doc assembler
                pipe.components.insert(insert_idx + 1, sentence_detector)
            return PipeUtils.configure_component_output_levels_to_sentence(pipe)
        elif new_output_level == 'document':
            return PipeUtils.configure_component_output_levels_to_document(pipe)

    @staticmethod
    def check_if_component_is_in_pipe(pipe: NLUPipeline, component_name_to_check, check_strong=True):
        """Check if a component_to_resolve with a given name is already in a component_list """
        for c in pipe.components:
            if check_strong and component_name_to_check == c.info.name:
                return True
            elif not check_strong and component_name_to_check in c.info.name:
                return True
        return False

    @staticmethod
    def check_if_there_component_with_col_in_components(component_list, features, except_component):
        """For a given list of features and a list of components, see if there are components taht provide this feature
        If yes, True, otherwise False
        """
        for c in component_list:
            if c.out_types[0] != except_component.out_types[0]:
                for f in ComponentUtils.clean_irrelevant_features(c.info.spark_output_column_names, True):
                    if f in features: return True

        return False

    @staticmethod
    def is_leaf_node(c, pipe: NLUPipeline) -> bool:
        """Check if a component_to_resolve is a leaf in the DAG.
        We verify by checking if any other_c is feeding from os_components.
        If yes, it is not a leaf. If nobody feeds from os_components, it's a leaf.
        """
        inputs = c.info.inputs
        for other_c in pipe.components:
            if c is not other_c:
                for f in other_c.info.inputs: 1

        return False

    @staticmethod
    def clean_AT_storage_refs(pipe: NLUPipeline):
        """Removes AT notation from all columns. Useful to reset component_list back to default state"""
        for c in pipe.components:
            if c.info.loaded_from_pretrained_pipe:
                continue
            c.info.inputs = [f.split('@')[0] for f in c.info.inputs]
            c.out_types = [f.split('@')[0] for f in c.out_types]
            c.info.spark_input_column_names = [f.split('@')[0] for f in c.info.spark_input_column_names]
            c.info.spark_output_column_names = [f.split('@')[0] for f in c.info.spark_output_column_names]
            c.info.spark_input_column_names = c.info.inputs.copy()
            c.info.spark_output_column_names = c.out_types.copy()

        return pipe

    @staticmethod
    def rename_duplicate_cols(pipe: NLUPipeline):
        """Rename cols with duplicate names"""
        for i, c in enumerate(pipe.components):
            for other_c in pipe.components:
                if c is other_c:
                    continue
                if c.loaded_from_pretrained_pipe:
                    continue
                if c.spark_output_column_names[0] == other_c.spark_output_column_names[0]:
                    c.spark_output_column_names[0] = f'{c.spark_output_column_names[0]}_{str(i)}'
        return pipe

    @staticmethod
    def find_trainable_embed_consumer(pipe: NLUPipeline):
        """Find traianble component_to_resolve which consumes emeddings.
        Returns index of component_to_resolve and type of embedding if found, otherwise returns -1 and None"""
        for i, c in enumerate(pipe.components):
            if c.trainable and c.has_storage_ref:
                return pipe.components.index(c), ComponentUtils.extract_embed_col(c, 'input')

        return -1, None

    @staticmethod
    def remove_convertable_storage_refs(required_features_ref, conversion_candidates, provided_features_ref):
        """Remove required storage ref features if conversion candidate has it, so that storage ref provider will not
        be downloaded twice """
        if len(conversion_candidates) == 0:
            return required_features_ref, conversion_candidates
        # ComponentUtils.extract_storage_ref_AT_notation_for_embeds
        for candidate in conversion_candidates:
            # candidate_at_storage_ref_feature = ComponentUtils.extract_storage_ref_AT_notation_for_embeds(
            #     candidate.component_candidate, 'output')
            if candidate.component_candidate is None:
                continue
            for feature in required_features_ref:
                # if feature not in provided_features_ref: # TODO revisit this after deep test
                #     # Feature not yet manifested by creating corresponding anno
                #     # Unless its also a storage ref candidate. In this scenario, the Feature is manifested but the Converter is missing.
                #     # Remove the feature from requirements, since its already there and will otherwise cause storage ref resolution to manifest again
                #     continue
                required_storage_ref = feature.split('@')[-1]
                if required_storage_ref == candidate.storage_ref:  # or candidate_at_storage_ref_feature == feature
                    # The feature is already provided, but not converted. We can remove it
                    required_features_ref.remove(feature)

        return required_features_ref, conversion_candidates

    @staticmethod
    def update_converter_storage_refs_and_cols(pipe: NLUPipeline, provided_features_ref, required_features_ref):
        """Storage ref of converters is initially empty string, i.e. '' .
        This method checks if  any convertable embeddings are provided, if yes it will update storage ref of converter
        , update the input/output columns with colname@storage_ref notation and mark it as resolved
        by removing it from the corrosponding lists"""

        for c in pipe.components:
            if c.name in [NLP_NODE_IDS.SENTENCE_EMBEDDINGS_CONVERTER, NLP_NODE_IDS.CHUNK_EMBEDDINGS_CONVERTER]:
                # Check if there are candidates that feed the converter, any word Embedding will work
                if c.storage_ref != '':
                    # If storage_ref is not '' then this is converter is already fixed, nothing to do
                    continue
                for other_c in pipe.components:
                    if other_c.has_storage_ref and other_c.type == AnnoTypes.TOKEN_EMBEDDING:
                        # Get original embed cols
                        in_embed = ComponentUtils.extract_embed_col(c, 'input')
                        out_embed = ComponentUtils.extract_embed_col(c, 'output')

                        if len(in_embed.split('@')) == 2:
                            # Storage ref is already on annotator, we dont ned to fix this
                            continue

                        c.spark_output_column_names.remove(out_embed)
                        c.spark_input_column_names.remove(in_embed)
                        provided_features_ref.remove(out_embed + '@')
                        required_features_ref.remove(in_embed + '@')
                        storage_ref = StorageRefUtils.extract_storage_ref(other_c)
                        in_embed = in_embed + '@' + storage_ref
                        out_embed = out_embed + '@' + storage_ref
                        c.spark_output_column_names.append(out_embed)
                        c.spark_input_column_names.append(in_embed)
                        provided_features_ref.append(out_embed)
                        required_features_ref.append(in_embed)
                        c.storage_ref = storage_ref

        return provided_features_ref, required_features_ref

    @staticmethod
    def add_metadata_to_pipe(pipe: NLUPipeline):
        """Write metadata fields to pipeline, for now only whether it contains
        OCR components or not. To be extended in the future
        """
        py_class_to_anno_id = AnnoClassRef.get_ocr_pyclass_2_anno_id_dict()

        for c in pipe.components:
            # Check for OCR componments
            if c.prefer_light_pipe:
                pipe.prefer_light = True

            if c.jsl_anno_py_class in py_class_to_anno_id.keys() or c.is_visual_annotator:
                pipe.contains_ocr_components = True
                if c.requires_image_format:
                    pipe.requires_image_format = True
                if c.requires_binary_format:
                    pipe.requires_binary_format = True

            # Check for licensed components
            if c.license in [Licenses.ocr, Licenses.hc]:
                pipe.has_licensed_components = True
            # Check for NLP Component, which is any open source
            if c.license == Licenses.open_source \
                    and c.name != NLP_NODE_IDS.WAV2VEC_FOR_CTC \
                    and c.name != NLP_NODE_IDS.HUBERT_FOR_CTC \
                    and c.name != NLP_NODE_IDS.WHISPER_FOR_CTC \
                    and c.name != NLP_NODE_IDS.AUDIO_ASSEMBLER:
                # TODO Table Assembler/VIT/ Other non txt open source
                pipe.has_nlp_components = True
            if c.type == AnnoTypes.QUESTION_TABLE_ANSWERER:
                pipe.has_table_qa_models = True

            # if c.type == AnnoTypes.CHUNK_MAPPER:
            #     pipe.prefer_light = True

            if c.type == AnnoTypes.QUESTION_SPAN_CLASSIFIER:
                pipe.has_span_classifiers = True

            if c.type == AnnoTypes.SPEECH_RECOGNIZER:
                pipe.contains_audio_components = True

            if c.type == AnnoTypes.IMAGE_CLASSIFICATION:
                pipe.contains_ocr_components = True
                pipe.has_nlp_components = False
            if c.jsl_anno_py_class == 'ImageAssembler':
                pipe.contains_ocr_components = True
            if c.is_light_pipe_incompatible:
                pipe.is_light_pipe_incompatible = True

        return pipe

    @staticmethod
    def replace_untrained_component_with_trained(nlu_pipe: NLUPipeline, spark_transformer_pipe):
        """Write metadata fields to pipeline, for now only whether it contains
        OCR components or not. To be extended in the future
        :return:
        :param nlu_pipe: NLU pipeline, which contains one untrained component
        :param spark_transformer_pipe: Spark Pipeline which contains fitted component version of the untrained one
        :return: NLU pipeline component list, where untrained component is replaced with a trained one
        """

        # Go through NLU pip and find the untrained component and replace with the trained one
        for i, trainable_c in enumerate(nlu_pipe.components):
            if trainable_c.trainable:

                # Construct trained NLU component with the trained Spark Model
                if trainable_c.license == Licenses.open_source:
                    trained_class_name = AnnoClassRef.JSL_anno2_py_class[trainable_c.trained_mirror_anno]
                    untrained_class_name = AnnoClassRef.JSL_anno2_py_class[trainable_c.jsl_anno_class_id]
                    trained_model = PipeUtils.get_model_of_class_from_spark_pipe(spark_transformer_pipe,
                                                                                 trained_class_name)
                    trained_component = jsl_id_to_empty_component(trainable_c.trained_mirror_anno).set_metadata(
                        trained_model, trainable_c.trained_mirror_anno, trainable_c.trained_mirror_anno, nlu_pipe.lang,
                        False,
                        Licenses.open_source)
                elif trainable_c.license == Licenses.hc:
                    trained_class_name = AnnoClassRef.JSL_anno_HC_ref_2_py_class[trainable_c.trained_mirror_anno]
                    untrained_class_name = AnnoClassRef.JSL_anno_HC_ref_2_py_class[trainable_c.jsl_anno_class_id]
                    trained_model = PipeUtils.get_model_of_class_from_spark_pipe(spark_transformer_pipe,
                                                                                 trained_class_name)
                    trained_component = jsl_id_to_empty_component(trainable_c.trained_mirror_anno).set_metadata(
                        trained_model, trainable_c.trained_mirror_anno, trainable_c.trained_mirror_anno, nlu_pipe.lang,
                        False, Licenses.hc)

                # update col names on new model_anno_obj
                trained_component.spark_input_column_names = trainable_c.spark_input_column_names
                trained_component.spark_output_column_names = trainable_c.spark_output_column_names
                trained_component.model.setInputCols(trained_component.spark_input_column_names)
                trained_component.model.setOutputCol(trained_component.spark_output_column_names[0])

                # Replace component in pipe
                nlu_pipe.components.remove(trainable_c)
                # nlu_pipe.components.insert(i, trained_component)

                #  remove the component from the NlpuPipe dict Keys and add the trained one
                pipe_key_to_delete = None
                for k in nlu_pipe.keys():
                    if nlu_pipe[k].__class__.__name__ == untrained_class_name:
                        pipe_key_to_delete = k
                del nlu_pipe[pipe_key_to_delete]
                # TODOf NER or other trainable, make sure we ad at the right place!
                nlu_pipe.add(trained_component, idx=i)
        return nlu_pipe.components

    @staticmethod
    def get_model_of_class_from_spark_pipe(spark_transformer_pipe, class_name):
        for model in spark_transformer_pipe.stages:
            if model.__class__.name == class_name:
                return model
        raise ValueError(f"Could not find model_anno_obj of requested class = {class_name}")

    @staticmethod
    def contains_t5_or_gpt(pipe: NLUPipeline):
        return PipeUtils.has_component_with_id(pipe, [NLP_NODE_IDS.GPT2, NLP_NODE_IDS.T5_TRANSFORMER])

    @staticmethod
    def add_sentence_detector_to_pipe_if_required(pipe: NLUPipeline):
        """
        1. For Tabla-QA the Question Tapas Col should originate from a doc type
                                          -> doc_question -> sent_question    |
        (Context/Questions) -> Multi-Doc -                                    => TAPAS
                                          -> doc_context  -> assembled_table  |


        right after the Multi-Doc-assembler we add a sentence Detector.
        Sentence Detectors Input is doc_question
        and we update TAPAS to take sent_question instead of doc_question
        :param pipe:
        """
        if not pipe.has_table_qa_models:
            return pipe
        PipeUtils.has_sentence_detector(pipe)

        # Create Sentence Detector & Set inputs to Document_question
        sent_detector = ComponentUniverse.components[NLP_NODE_IDS.SENTENCE_DETECTOR_DL]()
        sent_detector.set_metadata(sent_detector.get_default_model(), 'detect_sentence',
                                   'sentence_detector_dl', 'en', False, Licenses.open_source)
        sent_detector.set_input(str(NLP_FEATURES.DOCUMENT_QUESTION))
        # Insert Sentence Detector right after Multi-Doc
        multi_doc_idx = PipeUtils.get_component_idx_by_id(pipe, NLP_NODE_IDS.MULTI_DOCUMENT_ASSEMBLER)
        pipe.components.insert(multi_doc_idx + 1, sent_detector)
        # Update Tapas to use sentence_detector_question instead of doc_quesiton
        pipe.components[PipeUtils.get_component_idx_by_id(pipe, NLP_NODE_IDS.TAPAS_FOR_QA)].set_input(
            [str(NLP_FEATURES.ASSEMBLED_TABULAR_DATA), str(NLP_FEATURES.SENTENCE)])
        return pipe



import logging
from typing import List

from nlu.pipe.nlu_component import NluComponent
from nlu.pipe.utils.resolution.storage_ref_utils import StorageRefUtils
from nlu.universe.feature_node_ids import NLP_NODE_IDS, NLP_HC_NODE_IDS
from nlu.universe.feature_universes import NLP_FEATURES, OCR_FEATURES
from nlu.universe.logic_universes import AnnoTypes

logger = logging.getLogger('nlu')


class ComponentUtils:
    """Component and Column Level logic operations and utils"""

    @staticmethod
    def config_chunk_embed_converter(converter: NluComponent) -> NluComponent:
        '''For a Chunk to be added to a pipeline, configure its input/output and set storage ref to amtch the storage ref and
        enfore storage ref notation. This will be used to infer backward later which component_to_resolve should feed this consumer'''
        storage_ref = StorageRefUtils.extract_storage_ref(converter)
        input_embed_col = ComponentUtils.extract_embed_col(converter)
        new_embed_col_with_AT_notation = input_embed_col + "@" + storage_ref
        converter.info.inputs.remove(input_embed_col)
        converter.info.inputs.append(new_embed_col_with_AT_notation)
        converter.info.spark_input_column_names.remove(input_embed_col)
        converter.info.spark_input_column_names.append(new_embed_col_with_AT_notation)
        converter.model.setInputCols(converter.info.inputs)

        return converter

    @staticmethod
    def clean_irrelevant_features(feature_list, remove_AT_notation=False, remove_text=True):
        '''
        Remove irrelevant features from a list of component_to_resolve features
        Also remove the @notation from names, since they are irrelevant for ordering
        :param feature_list: list of features
        :param remove_AT_notation: remove AT notation from os_components names if true. Used for sorting
        :return: list with only relevant feature names
        '''
        # remove irrelevant missing features for pretrained models
        # Most of these should be provided externally by the user and cannot be resolved
        if 'text' in feature_list and remove_text:
            feature_list.remove('text')
        if 'raw_text' in feature_list:
            feature_list.remove('raw_text')
        if 'raw_texts' in feature_list:
            feature_list.remove('raw_texts')
        if 'label' in feature_list:
            feature_list.remove('label')
        if 'sentiment_label' in feature_list:
            feature_list.remove('sentiment_label')
        if '%%%feature_elements%%%' in feature_list:
            feature_list.remove('%%%feature_elements%%%')
        if NLP_FEATURES.RAW_AUDIO in feature_list:
            feature_list.remove(NLP_FEATURES.RAW_AUDIO)
        if NLP_FEATURES.ANY in feature_list:
            feature_list.remove(NLP_FEATURES.ANY)
        if NLP_FEATURES.RAW_QUESTION in feature_list:
            feature_list.remove(NLP_FEATURES.RAW_QUESTION)
        if NLP_FEATURES.RAW_QUESTION_CONTEXT in feature_list:
            feature_list.remove(NLP_FEATURES.RAW_QUESTION_CONTEXT)
        if NLP_FEATURES.SPARK_NLP_FILE_PATH in feature_list:
            feature_list.remove(NLP_FEATURES.SPARK_NLP_FILE_PATH)
        if NLP_FEATURES.SPARK_NLP_IMAGE in feature_list:
            feature_list.remove(NLP_FEATURES.SPARK_NLP_IMAGE)

        if OCR_FEATURES.BINARY_IMG in feature_list:
            feature_list.remove(OCR_FEATURES.BINARY_IMG)
        if OCR_FEATURES.BINARY_IMG in feature_list:
            feature_list.remove(OCR_FEATURES.BINARY_IMG)
        if OCR_FEATURES.FILE_PATH in feature_list:
            feature_list.remove(OCR_FEATURES.FILE_PATH)
        if OCR_FEATURES.BINARY_DOCX in feature_list:
            feature_list.remove(OCR_FEATURES.BINARY_DOCX)
        if OCR_FEATURES.BINARY_PDF in feature_list:
            feature_list.remove(OCR_FEATURES.BINARY_PDF)

        if remove_AT_notation:
            new_cs = []
            for c in feature_list:
                new_cs.append(c.split("@")[0])
            return new_cs
        return feature_list

    @staticmethod
    def component_has_embeddings_requirement(component: NluComponent):
        '''
        Check for the input component_to_resolve, wether it depends on some embedding. Returns True if yes, otherwise False.
        :param component:  The component_to_resolve to check
        :return: True if the component_to_resolve needs some specifc embedding (i.e.glove, bert, elmo etc..). Otherwise returns False
        '''
        return component.is_storage_ref_consumer

    @staticmethod
    def extract_storage_ref_AT_notation_for_embeds(component: NluComponent, col='input'):
        '''
        Extract <col>_embed_col@storage_ref notation from a component_to_resolve if it has a storage ref, otherwise '
        :param component:  To extract notation from
        :cols component_to_resolve:  Wether to extract for the input or output col
        :return: '' if no storage_ref, <col>_embed_col@storage_ref otherwise
        '''
        if col == 'input':
            e_col = next(filter(lambda s: 'embed' in s, component.spark_input_column_names))
        elif col == 'output':
            e_col = next(filter(lambda s: 'embed' in s, component.spark_output_column_names))
        stor_ref = StorageRefUtils.extract_storage_ref(component)
        return e_col + '@' + stor_ref

    @staticmethod
    def is_embedding_provider(component: NluComponent) -> bool:
        """Check if a NLU Component returns/generates embeddings """
        return component.is_storage_ref_producer

    @staticmethod
    def is_embedding_consumer(component: NluComponent) -> bool:
        """Check if a NLU Component consumes embeddings """
        return component.is_storage_ref_consumer

    @staticmethod
    def is_embedding_converter(component: NluComponent) -> bool:
        """Check if NLU component_to_resolve is embedding converter """
        return component.name in [NLP_NODE_IDS.SENTENCE_EMBEDDINGS_CONVERTER,
                                  NLP_NODE_IDS.SENTENCE_EMBEDDINGS_CONVERTER]

    @staticmethod
    def is_NER_provider(component: NluComponent) -> bool:
        """Check if a NLU Component wraps a NER/NER-Medical model_anno_obj """

        if component.name in [NLP_HC_NODE_IDS.MEDICAL_NER, NLP_HC_NODE_IDS.TRAINABLE_MEDICAL_NER, NLP_NODE_IDS.NER_DL,
                              NLP_NODE_IDS.TRAINABLE_NER_DL, NLP_NODE_IDS.TRAINABLE_NER_CRF,
                              NLP_NODE_IDS.NER_CRF, NLP_HC_NODE_IDS.ZERO_SHOT_NER]: return True
        if component.type == AnnoTypes.TRANSFORMER_TOKEN_CLASSIFIER: return True

    @staticmethod
    def is_NER_IOB_token_classifier(component: NluComponent) -> bool:
        """Check if a Token Classifier uses IOB PRediction format"""
        if not hasattr(component.model, 'getClasses'):
            return False
        return any(['-' in label for label in component.model.getClasses()])

    @staticmethod
    def is_NER_converter(component: NluComponent) -> bool:
        """Check if a NLU Component wraps a NER-IOB to NER-Pr etty converter """
        return component.name in [NLP_HC_NODE_IDS.NER_CONVERTER_INTERNAL, NLP_NODE_IDS.NER_CONVERTER]

    @staticmethod
    def extract_NER_col(component: NluComponent, column='input') -> str:
        """Extract the exact name of the NER IOB column in the component_to_resolve"""
        if column == 'input':
            for f in component.in_types:
                if f in [NLP_FEATURES.NAMED_ENTITY_IOB, NLP_FEATURES.TOKEN_CLASSIFICATION]:
                    return f
        if column == 'output':
            for f in component.out_types:
                if f in [NLP_FEATURES.NAMED_ENTITY_IOB, NLP_FEATURES.TOKEN_CLASSIFICATION]:
                    return f
        raise ValueError(f"Could not find NER col for component_to_resolve ={component}")

    @staticmethod
    def extract_NER_converter_col(component: NluComponent, column='input') -> str:
        """Extract the exact name of the NER-converter  column in the component_to_resolve"""
        if column == 'input':
            for f in component.in_types:
                if f == NLP_FEATURES.NAMED_ENTITY_IOB:
                    return f
        if column == 'output':
            for f in component.out_types:
                if f == NLP_FEATURES.NAMED_ENTITY_CONVERTED:
                    return f
        raise ValueError(f"Could not find NER Converter col for component_to_resolve ={component}")

    @staticmethod
    def extract_embed_col(component: NluComponent, column='input') -> str:
        """Extract the exact name of the embed column in the component_to_resolve"""
        if column == 'input':
            for c in component.spark_input_column_names:
                if 'embed' in c: return c
        if column == 'output':
            for c in component.spark_output_column_names:
                if 'embed' in c: return c
        raise ValueError(f"Could not find Embed col for component_to_resolve ={component}")

    @staticmethod
    def set_storage_ref_attribute_of_embedding_converters(pipe_list: List[NluComponent]):
        """For every embedding converter, we set storage ref attr on it, based on what the storage ref from it's provider is """
        for converter in pipe_list:
            if ComponentUtils.is_embedding_provider(converter) and ComponentUtils.is_embedding_converter(converter):
                # First find the embed col of the converter
                embed_col = ComponentUtils.extract_embed_col(converter, column='input')
                for provider in pipe_list:
                    # Now find the Embedding generator that is feeding the converter
                    if embed_col in provider.spark_output_column_names:
                        converter.storage_ref = StorageRefUtils.nlp_extract_storage_ref_nlp_model(provider)
        return pipe_list

    @staticmethod
    def extract_embed_level_identity(component: NluComponent, col='input'):
        """Figure out if component_to_resolve feeds on chunk/sent aka doc/word emb for either nput or output cols"""
        if col == 'input':
            if any(filter(lambda s: 'document_embed' in s, component.info.inputs)): return 'document_embeddings'
            if any(filter(lambda s: 'sentence_embed' in s, component.info.inputs)): return 'sentence_embeddings'
            if any(filter(lambda s: 'chunk_embed' in s, component.info.inputs)): return 'chunk_embeddings'
            if any(filter(lambda s: 'token_embed' in s, component.info.inputs)): return 'token_embeddings'
        elif col == 'output':
            if any(filter(lambda s: 'document_embed' in s, component.out_types)): return 'document_embeddings'
            if any(filter(lambda s: 'sentence_embed' in s, component.out_types)): return 'sentence_embeddings'
            if any(filter(lambda s: 'chunk_embed' in s, component.out_types)): return 'chunk_embeddings'
            if any(filter(lambda s: 'token_embed' in s, component.out_types)): return 'token_embeddings'

    @staticmethod
    def are_producer_consumer_matches(e_consumer: NluComponent, e_provider: NluComponent) -> bool:
        """Check for embedding_consumer and embedding_producer if they match storage_ref and output level wise wise """
        if StorageRefUtils.extract_storage_ref(e_consumer) == StorageRefUtils.extract_storage_ref(e_provider):
            if ComponentUtils.extract_embed_level_identity(e_consumer,
                                                           'input') == ComponentUtils.extract_embed_level_identity(
                e_provider, 'output'):
                return True
            ## TODO FALL BACK FOR BAD MATCHES WHICH ACTUALLY MATCH-> consult name space
        return False

    @staticmethod
    def get_nlu_ref_identifier(component: NluComponent) -> str:
        """The tail of a NLU ref after splitting on '.' gives a unique identifier for NON-Aliased components
         If result is '' , model_anno_obj UID will be used as identifier
         """
        tail = component.nlu_ref.split('.')[-1].split('@')[-1]
        if tail == '':
            logger.warning(
                f"Could not deduct tail from component_to_resolve={component}. This is intended for CustomModelComponents used in offline mode")
            tail = str(component.model)
        return tail

    @staticmethod
    def remove_storage_ref_from_features(features: List[str]):
        """Clean storage ref from every str in list """
        return [f.split('@')[0] for f in features]
